(*
Warning: IMPORTANT WARNING NOTICE: CompCert 2.4 DOES NOT BELONG TO MICROSOFT - IT IS THIRD PARTY TECHNOLOGY. IT IS CLEARED ONLY FOR RESEARCH USE BY A SPECIFIC MSR RESEARCH TEAM. DO NOT USE IT UNTIL YOU HAVE CLEARED ITS USE FOR YOUR PROJECT WITH YOUR LEGAL CONTACT
*)

(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** The whole compiler and its proof of semantic preservation *)

Require Import Compiler.
Require Import String.
Import Errors.

Local Open Scope string_scope.

Definition transf_rtl_program (cse cp dc: bool) (f: RTL.program) : res RTL.program :=
   OK f
   @@ print (print_RTL 0)
   @@ time "Tail calls" Tailcall.transf_program
   @@ print (print_RTL 1)
  @@@ time "Inlining" Inlining.transf_program
   @@ print (print_RTL 2)
   @@ time "Renumbering" Renumber.transf_program
   @@ print (print_RTL 3)
  @@@ (if cp then fun p => OK p @@ time "Constant propagation" Constprop.transf_program
   @@ print (print_RTL 4) else fun p => OK p)
   @@ time "Renumbering" Renumber.transf_program
   @@ print (print_RTL 5)
  @@@ (if cse then fun p => time "CSE" CSE.transf_program p
   @@ print (print_RTL 6) else fun p => OK p)
  @@@ (if dc then fun p => time "Dead code" Deadcode.transf_program p
   @@ print (print_RTL 7) else fun p => OK p).

Definition transf_cminor_program (cse cp dc: bool) (p: Cminor.program) : res RTL.program :=
   OK p
   @@ print print_Cminor
  @@@ time "Instruction selection" Selection.sel_program
  @@@ time "RTL generation" RTLgen.transl_program
  @@@ (transf_rtl_program cse cp dc).

Definition transf_clight_program (cse cp dc: bool) (p: Clight.program) : res RTL.program :=
  OK p 
   @@ print print_Clight
  @@@ time "Simplification of locals" SimplLocals.transf_program
  @@@ time "C#minor generation" Cshmgen.transl_program
  @@@ time "Cminor generation" Cminorgen.transl_program
  @@@ (transf_cminor_program cse cp dc).

Definition transf_c_program (cse cp dc: bool) (p: Csyntax.program) : res RTL.program :=
  OK p 
  @@@ time "Clight generation" SimplExpr.transl_program
  @@@ (transf_clight_program cse cp dc).

(** * Semantic preservation *)

Require Import MoreRTL.
Import Coqlib Smallstep.

Require SimplExprproof.
Require SimplLocalsproof.
Require Cshmgenproof.
Require Cminorgenproof.
Require Selectionproof.
Require RTLgenproof.
Require Tailcallproof.
Require Inliningproof.
Require Renumberproof.
Require Constpropproof.
Require CSEproof.
Require Deadcodeproof.

(** We prove that the [transf_program] translations preserve semantics
  by constructing the following simulations:
- Forward simulations from [Cstrategy] / [Cminor] / [RTL] to [RTL]
  (composition of the forward simulations for each pass).
- Backward simulations for the same languages
  (derived from the forward simulation, using receptiveness of the source
  language and determinacy of [Asm]).
- Backward simulation from [Csem] to [Asm]
  (composition of two backward simulations).

These results establish the correctness of the whole compiler! *)

Theorem transf_rtl_program_correct:
  forall cse cp dc p tp,
  transf_rtl_program cse cp dc p = OK tp ->
  forward_simulation (RTL.semantics p) (RTL.semantics tp)
  * backward_simulation (RTL.semantics p) (RTL.semantics tp).
Proof.
  intros.
  assert (F: forward_simulation (RTL.semantics p) (RTL.semantics tp)).
  {
  unfold transf_rtl_program, time in H.
  repeat rewrite compose_print_identity in H.
  simpl in H.
  set (p1 := Tailcall.transf_program p) in *.
  destruct (Inlining.transf_program p1) as [p11|] eqn:?; simpl in H; try discriminate.
  set (p12 := Renumber.transf_program p11) in *.
  (destruct cp; [
    set (p2 := Constprop.transf_program p12) in *;
    rewrite !print_identity in H; simpl in H
  | set (p2 := p12) in *
    ]); simpl in *;
    set (p21 := Renumber.transf_program p2) in *;
    (destruct cse; [
      destruct (CSE.transf_program p21) as [p3|] eqn:?; simpl in H; try discriminate;
      repeat rewrite compose_print_identity in H
      | set (p3 := p21) in *
    ]); simpl in *;
  (destruct dc; [
    repeat rewrite compose_print_identity, print_identity in H;
    destruct (Deadcode.transf_program p3) as [p31|] eqn:?; simpl in H; try discriminate
      | set (p31 := p3) in *
    ]); simpl in *;
  inv H;
  repeat match goal with
  | |- forward_simulation _ (RTL.semantics ?tp) => unfold tp
  | H : Deadcode.transf_program _ = OK ?tp |- forward_simulation _ (RTL.semantics ?tp) =>
    eapply compose_forward_simulation;
    [ | apply Deadcodeproof.transf_program_correct; exact H ]
  | H : CSE.transf_program _ = OK ?tp |- forward_simulation _ (RTL.semantics ?tp) =>
    eapply compose_forward_simulation;
    [ | apply CSEproof.transf_program_correct; exact H ]
  | |- forward_simulation _ (RTL.semantics (Renumber.transf_program ?tp)) =>
       eapply compose_forward_simulation;
       [ | apply Renumberproof.transf_program_correct ]
  | |- forward_simulation _ (RTL.semantics (Constprop.transf_program ?tp)) =>
       eapply compose_forward_simulation;
       [ | apply Constpropproof.transf_program_correct ]
  | H : Inlining.transf_program _ = OK ?tp |- forward_simulation _ (RTL.semantics ?tp) =>
    eapply compose_forward_simulation;
    [ | apply Inliningproof.transf_program_correct; exact H ]
  | |- forward_simulation _ (RTL.semantics (Tailcall.transf_program ?tp)) =>
       apply Tailcallproof.transf_program_correct
  end.
  }
  split. auto. 
  apply forward_to_backward_simulation. auto. 
  apply RTL.semantics_receptive.
  apply MoreRTL.semantics_determinate.
Qed.

Theorem transf_cminor_program_correct:
  forall cse cp dc p tp,
  transf_cminor_program cse cp dc p = OK tp ->
  forward_simulation (Cminor.semantics p) (RTL.semantics tp)
  * backward_simulation (Cminor.semantics p) (RTL.semantics tp).
Proof.
  intros.
  assert (F: forward_simulation (Cminor.semantics p) (RTL.semantics tp)).
  unfold transf_cminor_program, time in H.
  repeat rewrite compose_print_identity in H.
  simpl in H. 
  destruct (Selection.sel_program p) as [p1|] eqn:?; simpl in H; try discriminate.
  destruct (RTLgen.transl_program p1) as [p2|] eqn:?; simpl in H; try discriminate.
  eapply compose_forward_simulation. apply Selectionproof.transf_program_correct. eauto.
  eapply compose_forward_simulation. apply RTLgenproof.transf_program_correct. eassumption.
  exact (fst (transf_rtl_program_correct _ _ _ _ _ H)).

  split. auto. 
  apply forward_to_backward_simulation. auto. 
  apply Cminor.semantics_receptive.
  apply MoreRTL.semantics_determinate.
Qed.

Theorem transf_clight_program_correct:
  forall cse cp dc p tp,
  transf_clight_program cse cp dc p = OK tp ->
  forward_simulation (Clight.semantics1 p) (RTL.semantics tp)
  * backward_simulation (Clight.semantics1 p) (RTL.semantics tp).
Proof.
  intros. 
  assert (F: forward_simulation (Clight.semantics1 p) (RTL.semantics tp)).
  revert H; unfold transf_clight_program, time; simpl.
  rewrite print_identity.
  caseEq (SimplLocals.transf_program p); simpl; try congruence; intros p0 EQ0.
  caseEq (Cshmgen.transl_program p0); simpl; try congruence; intros p1 EQ1.
  caseEq (Cminorgen.transl_program p1); simpl; try congruence; intros p2 EQ2.
  intros EQ3.
  eapply compose_forward_simulation. apply SimplLocalsproof.transf_program_correct. eauto.
  eapply compose_forward_simulation. apply Cshmgenproof.transl_program_correct. eauto.
  eapply compose_forward_simulation. apply Cminorgenproof.transl_program_correct. eauto.
  exact (fst (transf_cminor_program_correct _ _ _ _ _ EQ3)). 

  split. auto. 
  apply forward_to_backward_simulation. auto. 
  apply Clight.semantics_receptive.
  apply MoreRTL.semantics_determinate.
Qed.

Theorem transf_cstrategy_program_correct:
  forall cse cp dc p tp,
  transf_c_program cse cp dc p = OK tp ->
  forward_simulation (Cstrategy.semantics p) (RTL.semantics tp)
  * backward_simulation (atomic (Cstrategy.semantics p)) (RTL.semantics tp).
Proof.
  intros.
  assert (F: forward_simulation (Cstrategy.semantics p) (RTL.semantics tp)).
  revert H; unfold transf_c_program, time; simpl.
  caseEq (SimplExpr.transl_program p); simpl; try congruence; intros p0 EQ0.
  intros EQ1.
  eapply compose_forward_simulation. apply SimplExprproof.transl_program_correct. eauto.
  exact (fst (transf_clight_program_correct _ _ _ _ _ EQ1)). 

  split. auto. 
  apply forward_to_backward_simulation.
  apply factor_forward_simulation. auto. eapply sd_traces. eapply MoreRTL.semantics_determinate.
  apply atomic_receptive. apply Cstrategy.semantics_strongly_receptive.
  apply MoreRTL.semantics_determinate.
Qed.

Theorem transf_c_program_correct:
  forall cse cp dc p tp,
  transf_c_program cse cp dc p = OK tp ->
  backward_simulation (Csem.semantics p) (RTL.semantics tp).
Proof.
  intros. 
  apply compose_backward_simulation with (atomic (Cstrategy.semantics p)).
  eapply sd_traces; eapply MoreRTL.semantics_determinate.
  apply factor_backward_simulation. 
  apply Cstrategy.strategy_simulation.
  apply Csem.semantics_single_events.
  eapply ssr_well_behaved; eapply Cstrategy.semantics_strongly_receptive.
  exact (snd (transf_cstrategy_program_correct _ _ _ _ _ H)).
Qed.
