(*
Warning: IMPORTANT WARNING NOTICE: CompCert 2.4 DOES NOT BELONG TO MICROSOFT - IT IS THIRD PARTY TECHNOLOGY. IT IS CLEARED ONLY FOR RESEARCH USE BY A SPECIFIC MSR RESEARCH TEAM. DO NOT USE IT UNTIL YOU HAVE CLEARED ITS USE FOR YOUR PROJECT WITH YOUR LEGAL CONTACT
*)

(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

open Printf
open Camlcoq
open Clflags

let rec omega = Datatypes.S omega
type run_mode = Int32 | QAP
let option_run : run_mode option ref = ref None
let option_input_public : string ref = ref ""
let option_input_private: string ref = ref ""
let option_cse : bool ref = ref true
let option_cp : bool ref = ref true
let option_dc : bool ref = ref true

let string_of_run_mode =
  function
  | Int32 -> "eval"
  | QAP -> "qap"

let openInput (fn: string) =
  begin
  try
    let fd = open_in_bin fn in
    object
      method close = close_in fd
      method next_int = 
        try
          input_line fd |> int_of_string
        with
        | End_of_file ->
            failwith ("no next int in " ^ fn)
    end
  with _ ->
    (* no file, just return random numbers *)
    object
      method close = ()
      method next_int = Random.int 10
    end
  end

(* Location of the compatibility library *)

let stdlib_path = ref(
  try
    Sys.getenv "COMPCERT_LIBRARY"
  with Not_found ->
    Configuration.stdlib_path)

let command cmd =
  if !option_v then begin
    prerr_string "+ "; prerr_string cmd; prerr_endline ""
  end;
  Sys.command cmd

let quote_options opts =
  String.concat " " (List.rev_map Filename.quote opts)

let quote_arguments args =
  String.concat " " (List.map Filename.quote args)

let safe_remove file =
  try Sys.remove file with Sys_error _ -> ()

(* Printing of error messages *)

let print_error oc msg =
  let print_one_error = function
  | Errors.MSG s -> output_string oc (camlstring_of_coqstring s)
  | Errors.CTX i -> output_string oc (extern_atom i)
  | Errors.POS i -> fprintf oc "%ld" (P.to_int32 i)
  in
    List.iter print_one_error msg;
    output_char oc '\n'

(* Determine names for output files.  We use -o option if specified
   and if this is the final destination file (not a dump file).
   Otherwise, we generate a file in the current directory. *)

let output_filename ?(final = false) source_file source_suffix output_suffix =
  match !option_o with
  | Some file when final -> file
  | _ -> 
    Filename.basename (Filename.chop_suffix source_file source_suffix)
    ^ output_suffix

(* A variant of [output_filename] where the default output name is fixed *)

let output_filename_default default_file =
  match !option_o with
  | Some file -> file
  | None -> default_file

(* From C to preprocessed C *)

let preprocess ifile ofile =
  let output =
    if ofile = "-" then "" else sprintf "> %s" ofile in
  let cmd =
    sprintf "%s -D__COMPCERT__ %s %s %s %s"
      Configuration.prepro
      (if Configuration.has_runtime_lib
       then sprintf "-I%s" !stdlib_path
       else "")
      (quote_options !prepro_options)
      ifile output in
  if command cmd <> 0 then begin
    if ofile <> "-" then safe_remove ofile;
    eprintf "Error during preprocessing.\n";
    exit 2
  end

(* From preprocessed C to Csyntax *)

let parse_c_file sourcename ifile =
  Sections.initialize();
  (* Simplification options *)
  let simplifs =
    "b" (* blocks: mandatory *)
(*  ^ (if !option_flonglong then "l" else "") *)
  ^ (if !option_fstruct_return then "s" else "")
  ^ (if !option_fbitfields then "f" else "")
  ^ (if !option_fpacked_structs then "p" else "")
  in
  (* Parsing and production of a simplified C AST *)
  let ast =
    match Parse.preprocessed_file simplifs sourcename ifile with
    | None -> exit 2
    | Some p -> p in
  (* Save C AST if requested *)
  if !option_dparse then begin
    let ofile = output_filename sourcename ".c" ".parsed.c" in
    let oc = open_out ofile in
    Cprint.program (Format.formatter_of_out_channel oc) ast;
    close_out oc
  end;
  (* Conversion to Csyntax *)
  let csyntax =
    match C2C.convertProgram ast with
    | None -> exit 2
    | Some p -> p in
  flush stderr;
  (* Save CompCert C AST if requested *)
  if !option_dcmedium then begin
    let ofile = output_filename sourcename ".c" ".compcert.c" in
    let oc = open_out ofile in
    PrintCsyntax.print_program (Format.formatter_of_out_channel oc) csyntax;
    close_out oc
  end;
  csyntax

(* Printing a behavior *)

let string_of_int i = i |> Integers.Int.unsigned |> Z.hex
let print_Z oc z = Printf.fprintf oc "%s" (Z.hex z)
let print_N oc n = print_Z oc (match n with N.N0 -> Z.Z0 | N.Npos p -> Z.Zpos p)

let string_of_constant =
  function
  | Expr.ConstInt i -> string_of_int i
  | Expr.ConstStackAddr ofs -> Printf.sprintf "stk[%s]" (string_of_int ofs)
  | Expr.ConstGlobal (s, ofs) -> Printf.sprintf "%s[%s]" (extern_atom s) (string_of_int ofs)

let print_val pv oc =
  function
  | AbStateImpl.Const c -> Printf.fprintf oc "%s" (string_of_constant c)
  | AbStateImpl.Var v -> Printf.fprintf oc "%a" pv v

let string_of_chunk =
  function
  | AST.Mint8signed -> "Mint8s"
  | AST.Mint8unsigned -> "Mint8u"
  | AST.Mint16signed -> "Mint16s"
  | AST.Mint16unsigned -> "Mint16u"
  | AST.Mint32 -> "Mint32"
  | AST.Mint64 -> "Mint64"
  | AST.Mfloat32 -> "Mfloat32"
  | AST.Mfloat64 -> "Mfloat64"
  | AST.Many32 -> "Many32"
  | AST.Many64 -> "Many64"

let string_of_list f ?sep:(s="; ") l =
  l |> List.map f |> String.concat s

let print_list (p: out_channel -> 'a -> unit) ?sep:(s="; ") (oc: out_channel) (l: 'a list): unit =
  l
  |> List.fold_left (fun b i -> Printf.fprintf oc "%s%a" (if b then s else "") p i; true) false
  |> (ignore : bool -> unit)

let print_event pv oc =
  function
  | AbState.AEinput (k, f, ofs, v) ->
      Printf.fprintf oc "vload %s %s[%s] → %a"
      (string_of_chunk k)
      (extern_atom f)
      (string_of_int ofs)
      (print_val pv) v
  | AbState.AEannot (msg, args) ->
      Printf.fprintf oc "annot %s(%a)"
      (extern_atom msg)
      (print_list (fun oc i -> Printf.fprintf oc "%s" (string_of_int i)) ~sep:", ") args
  | AbState.AEoutput (k, f, ofs, v) ->
      Printf.fprintf oc "vstore %s %s[%s] ← %a"
      (string_of_chunk k)
      (extern_atom f)
      (string_of_int ofs)
      (print_val pv) v

let print_trace oc pv tr : unit =
  tr |>
  List.iter (fun i -> Printf.fprintf oc "%a\n" (print_event pv) i)

let print_beh verbose pv p oc ((tr, r), s) =
  if verbose
  then (
    print_trace oc (pv (snd s)) tr;
    Printf.fprintf oc "Return: %a\n" (print_val (pv (snd s))) r
  );
  p s;
  ()

let string_of_Z_coeff z =
  if Z.compare Z.one z = 0
  then ""
  else Printf.sprintf "%s · " (Z.hex z)

let print_lc qap oc lc =
  let lc = match Maps.PTree.get lc (QAP.env qap) with
  | Some (Datatypes.Coq_inl (lc, rng)) -> lc
  | Some (Datatypes.Coq_inr br) -> failwith "print_lc: binary repr"
  | None -> assert false in
  if
    QAP.LC.fold
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
  (fun first x k ->
    Printf.fprintf oc "%s%s[%d]" (if first then "" else " + ") (string_of_Z_coeff (Z.of_N k)) (P.to_int x);
    false
    )
  lc
  true
  then Printf.fprintf oc "0"

let print_qap verbose qap =
  if verbose
  then (
  let z x = if x = "" then "0" else x in
  QAP.fold_equations
    (ModpProof.zpEquiv LargePrime.large_prime)
    (ModpProof.zpPlus LargePrime.large_prime)
    (ModpProof.zpZero LargePrime.large_prime)
  (fun v w y -> Printf.sprintf "(%s) × (%s) = %s" (z v) (z w) (z y))
  (fun acc x k ->
    Printf.sprintf "%s%s%s[%d]" acc (if acc = "" then "" else " + ")
    (string_of_Z_coeff (Z.of_N k))
    (P.to_int x))
  ""
  qap
  |> List.rev
  |> List.iter print_endline
  );
  let oc = open_out_bin "qap.bin" in
  QapIO.output_qap_bin oc qap;
  close_out oc

(* Binary decomposition of a ZP on [n] bits. [n] should be large enough. *)
let bits (f: bool -> 'a) (n0: int) (z: Modp.zp) (next: 'a list) : 'a list =
  let rec bits n z acc =
  BinNums.(
  begin match n, z with
  | 0, N0 -> List.rev_append acc next
  | 0, _ -> failwith (Printf.sprintf "bits: underflow %d" n0)
  | _, N0 -> bits (n - 1) z (f false :: acc)
  | _, Npos Coq_xH -> bits (n - 1) N0 (f true :: acc)
  | _, Npos (Coq_xO z') -> bits (n - 1) (Npos z') (f false :: acc)
  | _, Npos (Coq_xI z') -> bits (n - 1) (Npos z') (f true :: acc)
  end)
  in
  bits n0 z []

  (* From CompCert C AST to RTL *)

let compile_c_ast sourcename csyntax ofile =
  (* Prepare to dump Clight, RTL, if requested *)
  let set_dest dst opt ext =
    dst := if !opt then Some (output_filename sourcename ".c" ext)
    else None in
  set_dest PrintClight.destination option_dclight ".light.c";
  set_dest PrintCminor.destination option_dcminor ".cm";
  set_dest PrintRTL.destination option_drtl ".rtl";
  (* Run the interpreter *)
  let () =
    match !option_run with
  | Some m ->
      let () =
        Printf.printf "Interpreting (%s)\n" (string_of_run_mode m);
    let () = Random.self_init () in
    let compile_time_oracle () = {
      AbStateImpl.o_init =
        (open_out_bin "hints", openInput (!option_input_public));
      AbStateImpl.o_hint =
        (fun h (hints, pub) ->
          Printf.fprintf hints "%s\n"
          (match h with
           | AbStateImpl.Hint_Mult(true, _) -> "mul-save"
           | AbStateImpl.Hint_Mult(false, _) -> "mul-nosave"
           | AbStateImpl.Hint_BinDec(None, _) -> "bin-nodec"
           | AbStateImpl.Hint_BinDec(Some n, _) -> Printf.sprintf "bin-dec %d" (Nat.to_int n)
           | AbStateImpl.Hint_Assert(Some (_,n), _) -> Printf.sprintf "assert-save %d" (Nat.to_int n)
           | AbStateImpl.Hint_Assert(None, _) -> "assert-nosave"
          );
          Monads.STm.Result((), (hints, pub))
          );
      AbStateImpl.o_output =
        (fun v s ->
          ignore v;
          Monads.STm.Result((), s));
      AbStateImpl.o_input =
        (fun _ s ofs (hints, pub) ->
          let n = extern_atom s in
          if n = "input_compile_time"
          then
            let v = pub # next_int in
            Printf.fprintf hints "input:%d\n" v;
            Monads.STm.Result (AbInt.InputNow (true, Integers.Int.repr (Z.of_sint v)), (hints, pub))
    else if n = "input_public" || n = "input_private"
          then Monads.STm.Result (AbInt.InputLater, (hints, pub))
          else Monads.STm.Error ("wrong input" |> coqstring_of_camlstring |> Errors.msg));
    } in
    let run_time_oracle a = {
      AbStateImpl.o_init =
        (let hints = open_in_bin "hints" in
         let pub = openInput (!option_input_public) in
         let priv = openInput (!option_input_private) in
         (hints, pub, priv, []));
      AbStateImpl.o_input =
        (fun _ s ofs (hints, pub, priv, log) ->
          let n = extern_atom s in
          if n = "input_compile_time"
          then
            let h = input_line hints in
            let v = Scanf.sscanf h "input:%d" Z.of_sint in
            Monads.STm.Result (AbInt.InputNow (true, Integers.Int.repr v), (hints, pub, priv, log))
    else if n = "input_public" || n = "input_private"
          then
            let v = (if n = "input_private" then priv else pub) # next_int |> Z.of_sint |> Integers.Int.repr in
            match AbInt.Arith.of_int a v () with
            | Monads.STm.Result (lc, ()) ->
              let log = (AbStateImpl.Var lc) :: log in
              Monads.STm.Result (AbInt.InputNow (false, v), (hints, pub, priv, log))
            | Monads.STm.Error e -> Monads.STm.Error e
          else Monads.STm.Error ("wrong input" |> coqstring_of_camlstring |> Errors.msg));
      AbStateImpl.o_hint =
        (fun h (hints, pub, priv, log) ->
            let h' = input_line hints in
          let log =
          begin match h with
          | AbStateImpl.Hint_Mult(_, v) ->
            if h' = "mul-save" then v :: log
            else if h' = "mul-nosave" then log
            else (Printf.eprintf "bad hint: %s\n" h'; log)
          | AbStateImpl.Hint_BinDec(_, v) ->
            if h' = "bin-nodec" then log
            else Scanf.sscanf h' "bin-dec %d" (fun n ->
              bits (fun b -> AbStateImpl.Var (Modp.(if b then one else zero) LargePrime.large_prime))
              n
              v
              log
            )
          | AbStateImpl.Hint_Assert(_, v) ->
            if h' = "assert-nosave" then log
            else Scanf.sscanf h' "assert-save %d" (fun n ->
              AbStateImpl.Var (
                Modp.shift_inverse LargePrime.large_prime
                (Z.of_sint n) v
              ) :: log
            )
          end
          in
          Monads.STm.Result ((), (hints, pub, priv, log)));
      AbStateImpl.o_output =
        (fun v (hints, pub, priv, log) ->
          let log = v :: log in
          Monads.STm.Result ((), (hints, pub, priv, log)));
    } in
    match m with
    | Int32 ->
        let () =
          match QapGen.eval run_time_oracle !option_cse !option_cp !option_dc csyntax omega with
        | Monads.Result.Result beh ->
            beh |>
            print_beh true (fun _ -> print_N)
            (fun ((_, (hints, pub, priv, log)), _) ->
              close_in hints;
              pub # close;
              priv # close;
              let oc = open_out_bin "trace" in
              let log = AbStateImpl.Var (Modp.one LargePrime.large_prime):: List.rev log in
              List.iter (print_val QapIO.output_zp oc) log;
              close_out oc
              )
            stdout
        | Monads.Result.Error msg -> msg |> print_error stderr
              in ()
    | QAP ->
      let () =
        match QapGen.qap_gen (compile_time_oracle ()) !option_cse !option_cp !option_dc csyntax omega with
        | Monads.Result.Result beh ->
            beh |>
            print_beh !option_v (fun s oc lc -> print_lc s oc lc)
            (fun ((_, (hints, pub)), s) -> close_out hints; pub # close; Printf.printf "Final state:\n"; print_qap !option_v s)
            stdout
        | Monads.Result.Error msg -> msg |> print_error stderr
        in ()
    in ()
  | None ->
      "Use “-run [int|qap]” to run the interpreter."
      |> prerr_endline
  in
  (* Print Asm in text form
  let oc = open_out ofile in
  PrintRTL.print_program oc rtl;
  close_out oc
  *)
  ()


(* From C source to RTL *)

let compile_c_file sourcename ifile ofile =
  compile_c_ast sourcename (parse_c_file sourcename ifile) ofile

(* From Cminor to RTL *)

let compile_cminor_file ifile ofile =
  Sections.initialize();
  let ic = open_in ifile in
  let lb = Lexing.from_channel ic in
  try
    match RTLCompiler.transf_cminor_program !option_cse !option_cp !option_dc
            (CMtypecheck.type_program
              (CMparser.prog CMlexer.token lb)) with
    | Errors.Error msg ->
        print_error stderr msg;
        exit 2
    | Errors.OK p ->
        let oc = open_out ofile in
        PrintRTL.print_program oc p;
        close_out oc
  with Parsing.Parse_error ->
         eprintf "File %s, character %d: Syntax error\n"
                 ifile (Lexing.lexeme_start lb);
         exit 2
     | CMlexer.Error msg ->  
         eprintf "File %s, character %d: %s\n"
                 ifile (Lexing.lexeme_start lb) msg;
         exit 2
     | CMtypecheck.Error msg ->
         eprintf "File %s, type-checking error:\n%s"
                 ifile msg;
         exit 2

(* Processing of a .c file *)

let process_c_file sourcename =
  if !option_E then begin
    preprocess sourcename (output_filename_default "-");
    ""
  end else begin
    let preproname = Filename.temp_file "compcert" ".i" in
    preprocess sourcename preproname;
    if !option_interp then begin
      let csyntax = parse_c_file sourcename preproname in
      safe_remove preproname;
      Interp.execute csyntax;
      ""
    end else if !option_S then begin
      compile_c_file sourcename preproname
                     (output_filename ~final:true sourcename ".c" ".s");
      safe_remove preproname;
      ""
    end else begin
      let asmname = Filename.temp_file "compcert" ".s" in
      compile_c_file sourcename preproname asmname;
      safe_remove preproname;
      asmname
    end
  end

(* Processing of a .i / .p file (preprocessed C) *)

let process_i_file sourcename =
  if !option_interp then begin
    let csyntax = parse_c_file sourcename sourcename in
    Interp.execute csyntax;
    ""
  end else if !option_S then begin
    compile_c_file sourcename sourcename
                   (output_filename ~final:true sourcename ".c" ".s");
    ""
  end else begin
    let asmname = Filename.temp_file "compcert" ".s" in
    compile_c_file sourcename sourcename asmname;
    asmname
  end

(* Processing of a .cm file *)

let process_cminor_file sourcename =
  if !option_S then begin
    compile_cminor_file sourcename
                        (output_filename ~final:true sourcename ".cm" ".s");
    ""
  end else begin
    let asmname = Filename.temp_file "compcert" ".s" in
    compile_cminor_file sourcename asmname;
    asmname
  end

(* Record actions to be performed after parsing the command line *)

let actions : ((string -> string) * string) list ref = ref []

let push_action fn arg =
  actions := (fn, arg) :: !actions

let push_linker_arg arg =
  push_action (fun s -> s) arg

let perform_actions () =
  let rec perform = function
  | [] -> []
  | (fn, arg) :: rem -> let res = fn arg in res :: perform rem
  in perform (List.rev !actions)

(* Command-line parsing *)

let explode_comma_option s =
  match Str.split (Str.regexp ",") s with
  | [] -> assert false
  | hd :: tl -> tl

type action =
  | Set of bool ref
  | Unset of bool ref
  | Self of (string -> unit)
  | String of (string -> unit)
  | Integer of (int -> unit)

let rec find_action s = function
  | [] -> None
  | (re, act) :: rem ->
      if Str.string_match re s 0 then Some act else find_action s rem

let parse_cmdline spec usage =
  let acts = List.map (fun (pat, act) -> (Str.regexp pat, act)) spec in
  let rec parse i =
    if i < Array.length Sys.argv then begin
      let s = Sys.argv.(i) in
      match find_action s acts with
      | None ->
          if s <> "-help" && s <> "--help" 
          then eprintf "Unknown argument `%s'\n" s
          else printf "%s" usage;
          exit 2
      | Some(Set r) ->
          r := true; parse (i+1)
      | Some(Unset r) ->
          r := false; parse (i+1)
      | Some(Self fn) ->
          fn s; parse (i+1)
      | Some(String fn) ->
          if i + 1 < Array.length Sys.argv then begin
            fn Sys.argv.(i+1); parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; exit 2
          end
      | Some(Integer fn) ->
          if i + 1 < Array.length Sys.argv then begin
            let n =
              try
                int_of_string Sys.argv.(i+1)
              with Failure _ ->
                eprintf "Argument to option `%s' must be an integer\n" s;
                exit 2
            in
            fn n; parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; exit 2
          end
    end
  in parse 1

let usage_string =
"The CompCert C verified compiler, version " ^ Configuration.version ^ "
Usage: ccomp [options] <source files>
Recognized source files:
  .c             C source file
  .i or .p       C source file that should not be preprocessed
  .cm            Cminor source file
  .s             Assembly file
  .S             Assembly file that must be preprocessed
  .o             Object file
  .a             Library file
Processing options:
  -c             Compile to object file only (no linking), result in <file>.o
  -E             Preprocess only, send result to standard output
  -S             Compile to assembler only, save result in <file>.s
  -o <file>      Generate output in <file>
Preprocessing options:
  -I<dir>        Add <dir> to search path for #include files
  -D<symb>=<val> Define preprocessor symbol
  -U<symb>       Undefine preprocessor symbol
  -Wp,<opt>      Pass option <opt> to the preprocessor
Language support options (use -fno-<opt> to turn off -f<opt>) :
  -fbitfields    Emulate bit fields in structs [off]
  -flongdouble   Treat 'long double' as 'double' [off]
  -fstruct-return  Emulate returning structs and unions by value [off]
  -fvararg-calls Support calls to variable-argument functions [on]
  -funprototyped Support calls to old-style functions without prototypes [on]
  -fpacked-structs  Emulate packed structs [off]
  -finline-asm   Support inline 'asm' statements [off]
  -fall          Activate all language support options above
  -fnone         Turn off all language support options above
Debugging options:
  -g             Generate debugging information
Code generation options: (use -fno-<opt> to turn off -f<opt>) :
  -O             Optimize for speed [on by default]
  -Os            Optimize for code size
  -no-cse        Skip the cse phase
  -no-cp         Skip the constant-propagation phase
  -no-dc         Skip the deadcode-elim phase
  -ffpu          Use FP registers for some integer operations [on]
  -fsmall-data <n>  Set maximal size <n> for allocation in small data area
  -fsmall-const <n>  Set maximal size <n> for allocation in small constant area
  -ffloat-const-prop <n>  Control constant propagation of floats
                   (<n>=0: none, <n>=1: limited, <n>=2: full; default is full)
  -ftailcalls    Optimize function calls in tail position [on]
  -falign-functions <n>  Set alignment (in bytes) of function entry points
  -falign-branch-targets <n>  Set alignment (in bytes) of branch targets
  -falign-cond-branches <n>  Set alignment (in bytes) of conditional branches
Target processor options:
  -mthumb        (ARM only) Use Thumb2 instruction encoding
  -marm          (ARM only) Use classic ARM instruction encoding
Assembling options:
  -Wa,<opt>      Pass option <opt> to the assembler
Linking options:
  -l<lib>        Link library <lib>
  -L<dir>        Add <dir> to search path for libraries
  -Wl,<opt>      Pass option <opt> to the linker
Tracing options:
  -dparse        Save C file after parsing and elaboration in <file>.parse.c
  -dc            Save generated Compcert C in <file>.compcert.c
  -dclight       Save generated Clight in <file>.light.c
  -dcminor       Save generated Cminor in <file>.cm
  -drtl          Save RTL at various optimization points in <file>.rtl.<n>
General options:
  -stdlib <dir>  Set the path of the Compcert run-time library
  -v             Print external commands before invoking them
  -timings       Show the time spent in various compiler passes
Interpreter mode:
  -interp        Execute given .c files using the reference interpreter
  -quiet         Suppress diagnostic messages for the interpreter
  -trace         Have the interpreter produce a detailed trace of reductions
  -random        Randomize execution order
  -all           Simulate all possible execution orders

Pinocchio mode:
  -run [qap|int] Runs the RTL concrete interpreter
  -public  file  Use file as public input data (both compile-time and run-time)
  -private file  Use file as private input data (run-time only)
"

let language_support_options = [
  option_fbitfields; option_flongdouble;
  option_fstruct_return; option_fvararg_calls; option_funprototyped;
  option_fpacked_structs; option_finline_asm
]

let num_source_files = ref 0

let cmdline_actions =
  let f_opt name ref =
    ["-f" ^ name ^ "$", Set ref; "-fno-" ^ name ^ "$", Unset ref] in
  [
  "-I$", String(fun s -> prepro_options := s :: "-I" :: !prepro_options);
  "-D$", String(fun s -> prepro_options := s :: "-D" :: !prepro_options);
  "-U$", String(fun s -> prepro_options := s :: "-U" :: !prepro_options);
  "-[IDU].", Self(fun s -> prepro_options := s :: !prepro_options);
  "-[lL].", Self(fun s -> push_linker_arg s);
  "-o$", String(fun s -> option_o := Some s);
  "-E$", Set option_E;
  "-S$", Set option_S;
  "-c$", Set option_c;
  "-v$", Set option_v;
  "-g$", Self (fun s -> option_g := true; push_linker_arg s);
  "-stdlib$", String(fun s -> stdlib_path := s);
  "-dparse$", Set option_dparse;
  "-dc$", Set option_dcmedium;
  "-dclight$", Set option_dclight;
  "-dcminor", Set option_dcminor;
  "-drtl$", Set option_drtl;
  "-interp$", Set option_interp;
  "-quiet$", Self (fun _ -> Interp.trace := 0);
  "-trace$", Self (fun _ -> Interp.trace := 2);
  "-random$", Self (fun _ -> Interp.mode := Interp.Random);
  "-all$", Self (fun _ -> Interp.mode := Interp.All);
  "-run$", String(fun s -> option_run := if s = "qap" then Some QAP else Some Int32);
  "-public$", String( ( := ) option_input_public);
  "-private$", String( ( := ) option_input_private);
  ".*\\.c$", Self (fun s ->
      push_action process_c_file s;
      incr num_source_files);
  ".*\\.[ip]$", Self (fun s ->
      push_action process_i_file s;
      incr num_source_files);
  ".*\\.cm$", Self (fun s ->
      push_action process_cminor_file s;
      incr num_source_files);
  ".*\\.[oa]$", Self (fun s ->
      push_linker_arg s);
  "-Wp,", Self (fun s ->
      prepro_options := List.rev_append (explode_comma_option s) !prepro_options);
  "-Wa,", Self (fun s ->
      assembler_options := s :: !assembler_options);
  "-Wl,", Self (fun s ->
      push_linker_arg s);
  "-Os$", Set option_Osize;
  "-O$", Unset option_Osize;
  "-no-cse$", Unset option_cse;
  "-no-cp$", Unset option_cp;
  "-no-dc$", Unset option_dc;
  "-timings$", Set option_timings;
  "-mthumb$", Set option_mthumb;
  "-marm$", Unset option_mthumb;
  "-fsmall-data$", Integer(fun n -> option_small_data := n);
  "-fsmall-const$", Integer(fun n -> option_small_const := n);
  "-ffloat-const-prop$", Integer(fun n -> option_ffloatconstprop := n);
  "-falign-functions$", Integer(fun n -> option_falignfunctions := Some n);
  "-falign-branch-targets", Integer(fun n -> option_falignbranchtargets := n);
  "-falign-cond-branches", Integer(fun n -> option_faligncondbranchs := n);
  "-fall$", Self (fun _ ->
              List.iter (fun r -> r := true) language_support_options);
  "-fnone$", Self (fun _ ->
              List.iter (fun r -> r := false) language_support_options);
  ]
  @ f_opt "tailcalls" option_ftailcalls
  @ f_opt "longdouble" option_flongdouble
  @ f_opt "struct-return" option_fstruct_return
  @ f_opt "bitfields" option_fbitfields
  @ f_opt "vararg-calls" option_fvararg_calls
  @ f_opt "unprototyped" option_funprototyped
  @ f_opt "packed-structs" option_fpacked_structs
  @ f_opt "inline-asm" option_finline_asm
  @ f_opt "fpu" option_ffpu
  @ f_opt "sse" option_ffpu (* backward compatibility *)

let () =
  try
    Gc.set { (Gc.get()) with
                Gc.minor_heap_size = 524288; (* 512k *)
                Gc.major_heap_increment = 4194304 (* 4M *)
           };
    Printexc.record_backtrace true;
    Machine.config :=
      begin match Configuration.arch with
      | "powerpc" -> Machine.ppc_32_bigendian
      | "arm"     -> Machine.arm_littleendian
      | "ia32"    -> Machine.x86_32
      | _         -> assert false
      end;
    Builtins.set C2C.builtins;
    CPragmas.initialize();
    parse_cmdline cmdline_actions usage_string;
    let nolink = !option_c || !option_S || !option_E || !option_interp in
    if nolink && !option_o <> None && !num_source_files >= 2 then begin
      eprintf "Ambiguous '-o' option (multiple source files)\n";
      exit 2
    end;
    let linker_args = time "Total compilation time" perform_actions () in
    ignore linker_args
  with Sys_error msg ->
    eprintf "I/O error: %s.\n" msg; exit 2
