Require Import AbInt MoreLib.
Require modp.
Import Arith Monads.
Import String.
Import Integers.

Set Implicit Arguments.

Section ZP.
Import modp ZArith.
Import STm.Notations.
Import PP.PpsimplZ.

Local Open Scope Z_scope.

Variable p : positive.

Let of_int (i: int) : zp p := modp.const p (Z.to_N (Int.unsigned i)) (* FIXME: signedness ? *).

Let to_z (x: zp p) := Z.of_N x.
Let to_int (x: zp p) := Int.repr (to_z x).

Let overflow_check (z: Z) (err: unit → _) : STm.t (U.incr (S := unit)) Errors.errmsg unit :=
  if z <=? Zpos p - 1
  then U.ret tt
  else U.err (Errors.msg (err tt ++ "overflow")%string).

Existing Instance U.incr_trans.

Definition modp_arith: Arith.t Errors.errmsg (zp p) U.incr :=
  {| Arith.of_int i := U.ret (of_int i)
   ; Arith.to_int x := U.ret (to_int x)
   ; Arith.add i j :=
       do _ ← overflow_check (to_z i + to_z j) (λ _, "modp_arith.add: ")%string;
       U.ret (add i j)
   ; Arith.sub i j :=
       let res := sub i j in
       let expected := Int.sub (to_int i) (to_int j) in
       if Int.eq expected (to_int res) (* Run-time check of post-condition *)
       then U.ret res
       else U.ret (of_int expected)
   ; Arith.mul i j :=
       do _ ← overflow_check (to_z i * to_z j) (λ _, "modp_arith.mul: ")%string;
       U.ret (mul i j, false)
   ; Arith.land i j := U.ret (of_int (Int.and (to_int i) (to_int j)), (None, None))
   ; Arith.lor i j := U.ret (of_int (Int.or (to_int i) (to_int j)), (None, None))
   ; Arith.lxor i j := U.ret (of_int (Int.xor (to_int i) (to_int j)), (None, None))
   (* ; Arith.lnot i := U.ret (of_int (Int.not (to_int i)), None) *)
   (* ; Arith.shl i j := U.ret (of_int (Int.shl (to_int i) j)) *)
   ; Arith.shru i j := U.ret (of_int (Int.shru (to_int i) j), None)
   ; Arith.cmp s c i j :=
       match c with
       | Ceq => U.ret (of_int (if Int.eq (to_int i) (to_int j) then Int.one else Int.zero))
       | Cne =>U.ret (of_int ( if Int.eq (to_int i) (to_int j) then Int.zero else Int.one))
       | _ => U.err (Errors.msg "modp_arith: bad comparison")
       end
   ; Arith.init := tt
   ; Arith.input i r := match i with InputNow _ i => U.ret (of_int i) | InputLater => U.err (Errors.msg "runtime value") end
   ; Arith.output i := U.ret (i, None)
   ; Arith.assert i := if Int.eq (to_int i) Int.zero then U.err (Errors.msg "assertion failed") else U.ret None
   ; Arith.assert_eq i j := if Int.eq (to_int i) (to_int j) then U.ret tt else U.err (Errors.msg "assertion failed")
|}.

Hypothesis p_large : (Int.max_unsigned < Zpos p)%Z.

Lemma int_of_int (i: int) : to_int (of_int i) = i.
Proof.
  pose proof (Int.unsigned_range i).
  unfold to_int, of_int, to_z. simpl. rewrite N.mod_small. rewrite Z2N.id. rewrite Int.repr_unsigned. easy. Psatz.lia.
  apply N2Z.inj_lt. eapply Z.lt_le_trans. 2: apply Z.lt_pred_le; exact p_large. rewrite Z2N.id; easy.
Qed.

Lemma add_sound x y :
  to_z x + to_z y <=? Zpos p - 1 →
  Int.add (Int.repr (to_z x)) (Int.repr (to_z y)) = Int.repr (to_z (add x y)).
Proof.
  unfold add, to_z, is_true. intros LEB.
  simpl.
  apply overflow_elim.
  intros z [ ( H & <- ) | H ]. rewrite N2Z.inj_add. now rewrite add_repr.
  exfalso.
  clear -LEB H.
  ppsimpl. Psatz.lia.
Qed.

Lemma modp_arith_sound : Arith.spec modp_arith (λ _ ρ, ∀ x, ρ x = to_z x) (λ _, True).
Proof.
constructor.
- intros i σ x σ' Hincr H ρ H0 _. simpl in H. hsplit. rewrite H0. exact (conj (int_of_int i) I).
- intros x σ i σ' Hincr H ρ H0 _. simpl in H. hsplit. rewrite H0. exact (conj eq_refl I).
- simpl. unfold overflow_check.
  intros x y σ z σ' Hincr H ρ H0 _. flatten_eq hsplit. rewrite ! H0; clear H0. auto using add_sound.
- simpl.
  intros x y σ z σ' Hincr H ρ H0 _. simpl in *. flatten_eq hsplit. rewrite !H0; clear H0. auto.
  rewrite ! H0; clear H0. exact (conj (eq_sym (int_of_int _)) I).
- simpl. unfold overflow_check.
  intros x y σ z b σ' Hincr H ρ H0 _. flatten_eq hsplit. rewrite !H0; clear H0.
  unfold mul, to_z. simpl.
  rewrite mod_mul_spec.
  rewrite N.mod_small.
  rewrite N2Z.inj_mul. now rewrite mul_repr.
  unfold Interval.denote in *. subst to_z. unfold Interval.imin, Interval.imax in *. ppsimpl. Psatz.lia.
- simpl. intros x y σ z h σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.and (to_int x) (to_int y)))).
  rewrite int_of_int. reflexivity.
- simpl. intros x y σ z h σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.or (to_int x) (to_int y)))).
  rewrite int_of_int. reflexivity.
- simpl. intros x y σ z h σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.xor (to_int x) (to_int y)))).
  rewrite int_of_int. reflexivity.
  (*
- simpl. intros x σ z h σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.not (to_int x)))).
  rewrite int_of_int. reflexivity.
  *)
  (*
- simpl. intros x i σ z σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.shl (to_int x) i))).
  rewrite int_of_int. reflexivity.
  *)
- simpl. intros x i σ z h σ' Hincr H ρ H' _. refine (conj _ I).
  rewrite ! H'. clear H'. hsplit. subst.
  fold (to_int (of_int (Int.shru (to_int x) i))).
  rewrite int_of_int. reflexivity.
- simpl. intros s c x y σ z σ' Hincr H ρ H0 _. refine (conj _ I). rewrite ! H0. clear H0.
  flatten hsplit; destruct s; simpl; destruct (Int.eq _ _); symmetry; apply int_of_int.
- easy.
- easy.
- now simpl; intros; hsplit; subst.
- unfold assert. simpl. subst to_int. intros x y σ σ' Hincr H ρ H' _. rewrite H'. clear H'. flatten hsplit. auto.
- unfold assert_eq. simpl. subst to_int. intros x y σ σ' Hincr H ρ H' _. rewrite ! H'. clear H'. flatten hsplit. auto.
Qed.

End ZP.
