Require Import AbState.
Require Import AbInt.
Import String.
Import Relation_Definitions RelationClasses RelationPairs Morphisms.
Import MoreString MoreLib Monads Expr.
Import Globalenvs Memory Integers.

Import Maps Registers.
Import STm.Notations.

Unset Elimination Schemes.
Set Implicit Arguments.

Section VAL.

Context (ge: RTL.genv).

Context (var: Type).

Inductive val :=
| Const (c: constant)
| Var (v: var)
.

Let E := Errors.errmsg.
Context (S: Type) (sincr : relation S) `{sincr_refl: Reflexive _ sincr} `{sincr_trans: Transitive _ sincr}.

Variable arith : Arith.t E var sincr.

Context (Oracle: Type).

Inductive hint :=
(* The result of a multiplication needs to be saved *)
| Hint_Mult (b:bool) (r: val)
(* The variable [r] needs to be decomposed on [n] bits *)
| Hint_BinDec (n: option nat) (r: var)
(* Needs to produce the inverse of variable [r], shifted by [n×Int.modulus] *)
| Hint_Assert (n: option (var * nat)) (r: var)
.

Record oracle :=
{ o_input : AST.memory_chunk → AST.ident → int → STm.t (@U.incr Oracle) E input_data
; o_hint: hint → STm.t (@U.incr Oracle) E unit
; o_output: val → STm.t (@U.incr Oracle) E unit
; o_init: Oracle
}.

Variable o : oracle.

Section SA.

Record sa_state : Type :=
{
  (* Registers *)
  sa_reg: PTree.t val;
  (* Stack pointer *)
  sa_sp: Values.val;
  (* Stack frame; indices in the tree are (offsets)+1 *)
  sa_stk: PTree.t val;
  (* Global variables *)
  sa_global: AST.ident → int → result E val;
  (* Annotation about next public input value *)
  sa_next_range: option (int * int)
}.

Import BinPos.

Definition key_of_offset (ofs: int) : positive :=
  match Int.unsigned ofs with
  | Zpos p => Pos.succ p
  | _ => xH
  end.

Definition get_stk_err_msg (ofs: int): E :=
  Errors.msg ("get_stk at offset " ++ hex_of_Z (Int.signed ofs)).
Global Opaque get_stk_err_msg.

Definition get_stk (s: PTree.t val) (ofs: int) : result E val :=
  match s ! (key_of_offset ofs) with
  | Some v => Result v
  | None => Error (get_stk_err_msg ofs)
  end.

Let state := (sa_state * Oracle * S)%type.

Definition sa_incr : relation sa_state := λ x y, sa_sp x = sa_sp y.
Definition incr : relation state := (sa_incr * U.incr * sincr)%signature.

Instance sa_incr_refl : Reflexive sa_incr. Proof. easy. Defined.
Instance incr_refl : Reflexive incr. Proof. repeat apply RelProd_Reflexive; easy. Defined.
Instance sa_incr_trans : Transitive sa_incr. Proof. hnf. unfold sa_incr. congruence. Defined.
Instance incr_trans : Transitive incr. Proof. repeat apply RelProd_Transitive; try easy. exact sa_incr_trans. Defined.

Existing Instances U.incr_refl U.incr_trans.

Let get_reg (v: reg) : STm.t incr E val :=
  STm.res' (λ s, match (fst (fst s)).(sa_reg) ! v with Some r => Result r | None => Error (Errors.msg ("get_reg")) end).

Let put_reg (x: reg) (v: val) : STm.t incr E val :=
  STm.left (STm.left (STm.upd (λ s, (exist _ {| sa_reg := PTree.set x v s.(sa_reg) ; sa_sp := s.(sa_sp) ; sa_stk := s.(sa_stk) ; sa_global := s.(sa_global) ; sa_next_range := sa_next_range s |} eq_refl, v)))).

Let kill_reg (x: reg) : STm.t incr E unit :=
  STm.left (STm.left (STm.upd (λ s, (exist _ {| sa_reg := PTree.remove x s.(sa_reg) ; sa_sp := s.(sa_sp) ; sa_stk := s.(sa_stk) ; sa_global := s.(sa_global) ; sa_next_range := sa_next_range s |} eq_refl, tt)))).

Let put_stk (ofs: int) (v: val) : STm.t incr E unit :=
  STm.left (STm.left (STm.upd (λ s, (exist _ {| sa_reg := s.(sa_reg) ; sa_sp := s.(sa_sp) ; sa_stk := PTree.set (key_of_offset ofs) v s.(sa_stk) ; sa_global := sa_global s ; sa_next_range := sa_next_range s |} eq_refl, tt)))).

Let consume_next_range : STm.t incr E (option (int * int)) :=
  STm.left (STm.left (STm.upd (λ s, (exist _ {| sa_reg := sa_reg s ; sa_sp := sa_sp s ; sa_stk := sa_stk s ; sa_global := sa_global s ; sa_next_range := None |} eq_refl, sa_next_range s)))).

Definition two_consts (x x': val) : { c : _ & { c' | x = Const c ∧ x' = Const c' } }
  + { match x, x' with Const _, Const _ => False | _, _ => True end }.
refine
  match x, x' with
  | Const c, Const c' => inleft (existT _ c (exist _ c' (conj eq_refl eq_refl)))
  | _, _ => inright _
  end; exact I.
Defined.
Global Opaque two_consts.

Let is_neg a :=
  do tmp ← STm.right (
       do mask ← Arith.of_int arith (Int.repr Int.half_modulus);
       do a'' ← Arith.add arith a mask;
       do a' ← Arith.add arith a'' mask;
       Arith.is_neg arith a');
  let '(v, ((h₀,x₀), (h₁,x₁), (h₂,x₂))) := tmp in
  do _ ← STm.left (STm.right (
                       do _ ← o.(o_hint) (Hint_BinDec h₀ x₀);
                       do _ ← o.(o_hint) (Hint_BinDec h₁ x₁);
                       o.(o_hint) (Hint_BinDec h₂ x₂)));
    STm.ret v.

Definition is_slt0 op x y : { v | op = Bcmp true Clt ∧ x = Var v ∧ y = Const (ConstInt Int.zero) } + { True } :=
  match op with
  | Bcmp true Clt =>
    match x with
    | Var v =>
      match y with
      | Const (ConstInt i) =>
        (if Int.eq i Int.zero as b
            return (b = true → i = Int.zero) → { v | _ ∧ _ ∧ _ } + { True }
         then λ EQ, inleft (exist _ v (conj eq_refl (conj eq_refl (f_equal (λ x, Const (ConstInt x)) (EQ eq_refl))))) else λ _, inright I)
          (@int_eq_true i Int.zero)
      | _ => inright I
      end
    | _ => inright I
    end
  | _ => inright I
  end.
Global Opaque is_slt0.

Definition binop (op: binary_operator) (x y: val) : STm.t incr E val :=
  match two_consts x y with
  | inleft (existT c (exist d _)) =>
     do r ← STm.res (result_lift Const (binop_const op c d));
     STm.ret r
  | inright _ =>
  match is_slt0 op x y with
  | inleft (exist a _) =>
    do v ← is_neg a; STm.ret (Var v)
  | inright _ =>
  let with_arith (a: STm.t sincr E var) : STm.t incr E val :=
    STm.lift Var (STm.right a) in
  let with_mul (a: STm.t sincr E (var * bool)) : STm.t incr E val :=
    do r_b ← STm.right a;
    let '(r, b) := r_b in
    let r := Var r in
    do _ ← STm.left (STm.right (o.(o_hint) (Hint_Mult b r)));
    STm.ret r
  in
  let with_bin2 (a: STm.t sincr E (var * (option nat * option nat) * var * var)) : STm.t incr E val :=
    do r_h_i ← STm.right a;
    let '((r, (hx, hy)), x, y) := r_h_i in
    do _ ← STm.left (STm.right (o.(o_hint) (Hint_BinDec hx x)));
    do _ ← STm.left (STm.right (o.(o_hint) (Hint_BinDec hy y)));
    STm.ret (Var r)
  in
  let with_shru (a: STm.t sincr E ((var * option nat) * var)) : STm.t incr E val :=
    do r_h_i ← STm.right a;
    let '((r, hx), x) := r_h_i in
    do _ ← STm.left (STm.right (o.(o_hint) (Hint_BinDec hx x)));
    STm.ret (Var r)
  in
  match op, x, y with
  | Badd, Var v, Const (ConstInt i) => with_arith (do k ← Arith.of_int arith i; arith.(Arith.add) v k)
  | Badd, Const (ConstInt i), Var v => with_arith (do k ← Arith.of_int arith i; arith.(Arith.add) k v)
  | Badd, Var v, Var w => with_arith (Arith.add arith v w)
  | Bsub, Var v, Const (ConstInt i) => with_arith (do k ← Arith.of_int arith i; arith.(Arith.sub) v k)
  | Bsub, Const (ConstInt i), Var v => with_arith (do k ← Arith.of_int arith i; arith.(Arith.sub) k v)
  | Bsub, Var v, Var w => with_arith (Arith.sub arith v w)
  | Bcmp s c, Var v, Var w => with_arith (Arith.cmp arith s c v w)
  | Bcmp s c, Var v, Const (ConstInt i) => with_arith (do k ← Arith.of_int arith i; arith.(Arith.cmp) s c v k)
  | Bcmp s c, Const (ConstInt i), Var v => with_arith (do k ← Arith.of_int arith i; arith.(Arith.cmp) s c k v)
  | Bmul, Var v, Var w => with_mul (Arith.mul arith v w)
  | Bmul, Var v, Const (ConstInt i) => with_mul (do k ← Arith.of_int arith i; arith.(Arith.mul) v k)
  | Bmul, Const (ConstInt i), Var v => with_mul (do k ← Arith.of_int arith i; arith.(Arith.mul) k v)
  | Band, Var v, Const (ConstInt i) => with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.land arith v k;
      STm.ret (z, v, k)
    )
  | Band, Const (ConstInt i), Var w=> with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.land arith k w;
      STm.ret (z, k, w)
    )
  | Band, Var v, Var w => with_bin2 (
      do z ← Arith.land arith v w;
      STm.ret (z, v, w)
    )

  | Bor, Var v, Const (ConstInt i) => with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.lor arith v k;
      STm.ret (z, v, k)
    )
  | Bor, Const (ConstInt i), Var w=> with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.lor arith k w;
      STm.ret (z, k, w)
    )
  | Bor, Var v, Var w => with_bin2 (
      do z ← Arith.lor arith v w;
      STm.ret (z, v, w)
    )

  | Bxor, Var v, Const (ConstInt i) => with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.lxor arith v k;
      STm.ret (z, v, k)
    )
  | Bxor, Const (ConstInt i), Var w=> with_bin2 (
      do k ← Arith.of_int arith i;
      do z ← Arith.lxor arith k w;
      STm.ret (z, k, w)
    )
  | Bxor, Var v, Var w => with_bin2 (
      do z ← Arith.lxor arith v w;
      STm.ret (z, v, w)
    )

  | Bshl, Var v, Const (ConstInt i) => with_mul (
    do _ ← STm.assert (Sumbool.sumbool_of_bool (Int.ltu i Int.iwordsize))
      (λ _, Errors.msg ("shl too large: " ++ hex_of_int i)%string);
    (* TODO: use Arith.shl when available *)
    do k ← Arith.of_int arith (Int.shl Int.one i);
    Arith.mul arith v k
    )

  | Bshr false, Var v, Const (ConstInt i) => with_shru (
    do _ ← STm.assert (Sumbool.sumbool_of_bool (Int.ltu i Int.iwordsize))
      (λ _, Errors.msg ("shru too large: " ++ hex_of_int i)%string);
    do z ← Arith.shru arith v i;
    STm.ret (z, v)
    )

  | Bnot, Var v, Const (ConstInt i) => with_bin2 (
    (* TODO: use Arith.lnot when available *)
    do k ← Arith.of_int arith Int.mone;
    do z ← Arith.lxor arith v k;
    STm.ret (z, v, k)
    )

  | _, _, _ => STm.err _ (Errors.msg ("binop: " ++ string_of_binop op)%string)
  end
  end
  end.

Definition eval :=
  eval_expr
    get_reg Const
    (λ v, match v with
            | Const (ConstStackAddr ofs) => STm.left (STm.left (STm.res' (λ s, get_stk s.(sa_stk) ofs)))
            | Const (ConstGlobal gv ofs) => STm.left (STm.left (STm.res' (λ s, s.(sa_global) gv ofs)))
            | _ => STm.err _ (Errors.msg "bad load") end)
    binop.

Import List.

Definition sa : state_abstraction incr E val :=
{| sa_assign x e := do v ← eval e; do _ ← put_reg x v; STm.ret tt

 ; sa_forget := kill_reg

 ; sa_store x e :=
     do a ← eval x;
     do v ← eval e;
     match a with
     | Const (ConstStackAddr ofs) => put_stk ofs v
     | _ => STm.err _ (Errors.msg ("bad store to " ++ match a with Const c => string_of_constant c | Var _ => "variable" end))
     end

 ; sa_vload dst κ f ofs :=
     do n ← STm.left (STm.right (o.(o_input) κ f ofs));
     do r ← consume_next_range;
     do v ← STm.lift Var (STm.right (Arith.input arith n r));
     put_reg dst v

 ; sa_vstore dst κ f ofs arg :=
     do v ← get_reg arg;
     do x ←
        match v with
        | Const (ConstInt _) => STm.ret v
        | Var x =>
          do y_h ← STm.right (arith.(Arith.output) x);
          let '(y, h) := y_h in
          do _ ← STm.left (STm.right (o.(o_hint) (Hint_BinDec h x)));
          STm.ret (Var y)
        | Const _ => STm.err _ (Errors.msg "volatile store pointer")
        end;
     do _ ← kill_reg dst;
     do _ ← STm.left (STm.right (o.(o_output) x));
     STm.ret x

 ; sa_annot_input dst min max :=
     STm.left (STm.left (STm.upd (λ s, (exist _ {| sa_reg := PTree.remove dst (sa_reg s) ; sa_sp := sa_sp s ; sa_stk := sa_stk s ; sa_global := sa_global s ; sa_next_range := Some (min, max) |} eq_refl, tt))))

 ; sa_return e :=
     do v ← eval e;
     match v with
     | Const (ConstInt _) => STm.ret v
     | Const (ConstStackAddr _ | ConstGlobal _ _) => STm.err _ (Errors.msg "returned an address")
     | Var _ => STm.err _ (Errors.msg "TODO: return variable")
     end

 ; sa_init ge m₀ :=
    let stk := Mem.nextblock m₀ in
     (*
     let glob s (ofs: int) :=
         match assoc Pos.eqb s (AST.prog_defs prog) with
         | Some (AST.Gvar gv) => match AST.gvar_init gv with AST.Init_int32 i :: nil => Result (Const (ConstInt i))
                                                          | _ => Error (Errors.msg ("bad init list")) end
         | _ => Error (Errors.msg ("bad global"))
         end
     in
      *)
    let glob s (ofs: int) :=
        match Genv.find_symbol ge s with
        | None => Error (Errors.msg ("bad global"))
        |  Some b => match Mem.load AST.Mint32 m₀ b (Int.unsigned ofs) with
                   | None => Error (Errors.msg ("bad global load"))
                   | Some (Values.Vint i) => Result (Const (ConstInt i))
                   | Some _ => Error (Errors.msg ("wrong global"))
                     end
        end
    in
     Result ({| sa_reg := PTree.empty _ ; sa_sp := Values.Vptr stk Int.zero ; sa_stk := PTree.empty _ ; sa_global := glob ; sa_next_range := None |}, o.(o_init), Arith.init arith)

 ; sa_eval e :=
    do v ← eval e;
    match v with
    | Const (ConstInt i) => STm.ret (Values.Vint i)
    | Const (ConstGlobal g ofs) =>
      match Genv.find_symbol ge g with
      | Some b => STm.ret (Values.Vptr b ofs)
      | None => STm.err _ (Errors.msg ("sa_eval: unknown symbol"))
      end
    | Const c => STm.err _ (Errors.msg ("sa_eval: bad constant " ++ string_of_expr e ++ " → " ++ string_of_constant c))
    | Var x => do i ← STm.right (Arith.to_int arith x); STm.ret (Values.Vint i)
    end

 ; sa_assert e :=
 match e with
   | EBinop (Bcmp true Cne) (EVar r) (EConst (ConstInt i)) =>
     if Int.eq i Int.zero
     then
       do x ← get_reg r;
       match x with
       | Var v =>
           do q ← STm.right (Arith.assert arith v);
           STm.left (STm.right (o.(o_hint) (Hint_Assert q v)))
       | Const cst => STm.err _ (Errors.msg ("sa_assert: const " ++ string_of_constant cst))
       end
     else STm.err _ (Errors.msg ("sa_assert: bad expr"))
   | EBinop (Bcmp true Ceq) (EVar x) (EVar y) =>
     do u ← get_reg x;
     do v ← get_reg y;
       match u, v with
       | Var a, Var b => STm.right (Arith.assert_eq arith a b)
       | _, _ => STm.err _ (Errors.msg ("sa_assert: bad_expr again"))
       end
   | EBinop (Bcmp true Ceq) (EVar x) (EConst (ConstInt i)) =>
     do u ← get_reg x;
     match u with
     | Var a => STm.right (do b ← Arith.of_int arith i; Arith.assert_eq arith a b)
     | _ => STm.err _ (Errors.msg ("sa_assert: bad_expr again"))
     end
   | EBinop (Bcmp true Clt) (EVar x) (EConst (ConstInt i))
   | EBinop (Bcmp true Cgt) (EConst (ConstInt i)) (EVar x) =>
     if Int.eq i Int.zero
     then
     do u ← get_reg x;
     match u with
     | Var a =>
       do v ← is_neg a;
       do q ← STm.right (Arith.assert arith v);
       STm.left (STm.right (o.(o_hint) (Hint_Assert q v)))
     | _ => STm.err _ (Errors.msg ("sa_assert: bad signed comparison"))
     end
     else STm.err _ (Errors.msg ("sa_assert: signed comparison to a non-zero constant"))
   | _ => STm.err _ (Errors.msg ("sa_assert: TODO " ++ string_of_expr e))
 end

 ; sa_is_int s r :=
     match (fst (fst s)).(sa_reg) ! r with
     | Some v =>
       match v with
       | Const (ConstInt _) => true
       | Const (ConstStackAddr _ | ConstGlobal _ _) => false
       | Var x => true
       end
     | None => false
     end
|}.

End SA.

End VAL.

Arguments Const {var} c.
Arguments Var {var} v.
