Require Import Utf8.
Require Import Monads MoreString.
Require Import AST.
Import Coqlib Integers String Relation_Definitions RelationClasses.

Unset Elimination Schemes.
Set Implicit Arguments.

Section VAR.

Variable var : Type.

Inductive constant :=
| ConstInt (i: int)
| ConstStackAddr (ofs: int)
| ConstGlobal (s: ident) (ofs: int)
.

Definition string_of_constant (cst: constant) : string :=
  match cst with
  | ConstInt i => hex_of_int i
  | ConstStackAddr ofs => "stack(" ++ hex_of_int ofs ++ ")"
  | ConstGlobal s ofs => "global-" ++ hex_of_pos s ++ "("++ hex_of_int ofs ++")"
  end%string.

Inductive binary_operator :=
| Badd | Bsub | Bmul
| Bnot
| Band | Bor | Bxor
| Bshl | Bshr (signed: bool)
| Bcmp (signed: bool) (c: comparison).

Definition string_of_binop op :=
  match op with
  | Badd => "Badd"
  | Bsub => "Bsub"
  | Bmul => "Bmul"
  | Bnot => "Bnot"
  | Band => "Band"
  | Bor => "Bor"
  | Bxor => "Bxor"
  | Bshl => "Bshl"
  | Bshr s => "Bshr" ++ if s then "" else "u"
  | Bcmp s c => "Bcmp" ++ (if s then " signed " else " unsigned ") ++ string_of_comparison c
  end%string.

Let E := Errors.errmsg.

Definition binop_const (op: binary_operator) (x y: constant) : result E constant :=
  match op, x, y with
  | Badd, ConstInt i, ConstInt j => Result (ConstInt (Int.add i j))
  | Badd, ConstInt i, ConstStackAddr j
  | Badd, ConstStackAddr i, ConstInt j
    => Result (ConstStackAddr (Int.add i j))
  | Badd, ConstInt i, ConstGlobal s j
  | Badd, ConstGlobal s i, ConstInt j
    => Result (ConstGlobal s (Int.add i j))
  | Bsub, ConstInt i, ConstInt j => Result (ConstInt (Int.sub i j))
  | Bmul, ConstInt i, ConstInt j => Result (ConstInt (Int.mul i j))
  | Bnot, ConstInt i, _ => Result (ConstInt (Int.not i))
  | Band, ConstInt i, ConstInt j => Result (ConstInt (Int.and i j))
  | Bor, ConstInt i, ConstInt j => Result (ConstInt (Int.or i j))
  | Bxor, ConstInt i, ConstInt j => Result (ConstInt (Int.xor i j))
  | Bshl, ConstInt i, ConstInt j =>
    if Int.ltu j Int.iwordsize
    then Result (ConstInt (Int.shl i j))
    else Error (Errors.msg ("binop_const: overflow in " ++ string_of_binop op)%string)
  | Bshr s, ConstInt i, ConstInt j =>
    if Int.ltu j Int.iwordsize
    then Result (ConstInt ((if s then Int.shr else Int.shru) i j))
    else Error (Errors.msg ("binop_const: overflow in " ++ string_of_binop op)%string)
  | Bcmp s c, ConstInt i, ConstInt j => Result (ConstInt (if ((if s then Int.cmp else Int.cmpu) c i j) then Int.one else Int.zero))
  | _, _, _ => Error (Errors.msg ("binop_const: " ++ string_of_binop op)%string)
  end.

Local Set Elimination Schemes.
Inductive expr :=
| EVar (v: var)
| EConst (c: constant)
| ELoad (e: expr)
| EBinop (op: binary_operator) (l r: expr)
.

Definition string_of_expr : expr → string :=
  (expr_rect (λ _, string)
            (λ _, "var ?")
            string_of_constant
            (λ _ s, "*" ++ s)
            (λ op _ x _ y, "(" ++ x ++ " " ++ string_of_binop op ++ " " ++ y ++ ")")
  )%string.

End VAR.

Arguments EVar {var} _.
Arguments EConst {var} c.

Import STm.Notations.

Section EVAL.

Variables var E state val: Type.
Context (incr: relation state) `{Reflexive _ incr} `{Transitive _ incr}.

Variable get_var : var → STm.t incr E val.
Variable get_const : constant → val.
Variable load : val → STm.t incr E val.
Variable binop : binary_operator → val → val → STm.t incr E val.

Fixpoint eval_expr (e: expr var) :=
  match e with
  | EVar v => get_var v
  | EConst c => STm.ret (get_const c)
  | ELoad e =>
    do a ← eval_expr e;
    load a
  | EBinop op x y =>
    do x' ← eval_expr x;
    do y' ← eval_expr y;
    binop op x' y'
  end.

End EVAL.

Require RTL.
Import Registers Values Globalenvs Memory.

Section GE.

Context (F V: Type) (ge : Genv.t F V).

Definition eval_constant sp (c: constant) : Values.val :=
  match c with
  | ConstInt i => Values.Vint i
  | ConstStackAddr ofs => Values.Val.add sp (Values.Vint ofs)
  | ConstGlobal s ofs => match Genv.find_symbol ge s with Some b => Values.Vptr b ofs | None => Values.Vundef end
  end.

Existing Instances U.incr_refl U.incr_trans.
Definition rtl_eval_expr sp (rs: RTL.regset) (m: mem) : expr reg → STm.t (U.incr (S := unit)) _ Values.val :=
  eval_expr
    (λ r, U.ret (rs # r))
    (eval_constant sp)
    (λ a, match Mem.loadv AST.Mint32 m a with Some v => U.ret v | None => U.err (Errors.msg "bad load") end)
    (λ op x y, U.ret (
      match op with
      | Badd => Values.Val.add
      | Bsub => Values.Val.sub
      | Bmul => Values.Val.mul
      | Bnot => λ u v, Values.Val.notint u
      | Band => Values.Val.and
      | Bor => Values.Val.or
      | Bxor => Values.Val.xor
      | Bshl => Values.Val.shl
      | Bshr s => if s then Values.Val.shr else Values.Val.shru
      | Bcmp s c => (if s then Values.Val.cmp else Values.Val.cmpu (Mem.valid_pointer m)) c
      end x y)).

End GE.
