Require Import Utf8 String BinNums.
Require Import MoreLib.

Require Import RTL.
Import Coqlib AST Integers Op.

Local Open Scope string_scope.

Definition string_of_chunk (κ: memory_chunk) : string :=
  match κ with
  | Mint8signed => "Mint8signed"
  | Mint8unsigned => "Mint8unsigned"
  | Mint16signed => "Mint16signed"
  | Mint16unsigned => "Mint16unsigned"
  | Mint32 => "Mint32"
  | Mint64 => "Mint64"
  | Mfloat32 => "Mfloat32"
  | Mfloat64 => "Mfloat64"
  | Many32 => "Many32"
  | Many64 => "Many64"
  end.

Definition string_of_addressing (addr: addressing) : string :=
  match addr with
  | Aindexed _ => "indexed"
  | Aindexed2 _ => "indexed2"
  | Ascaled _ _ => "scaled"
  | Aindexed2scaled _ _ => "indexed2 scaled"
  | Aglobal _ _ => "global"
  | Abased _ _ => "based"
  | Abasedscaled _ _ _ => "based scaled"
  | Ainstack _ => "instack"
  end.

Definition string_of_comparison (c: comparison) : string :=
  match c with
  | Ceq => "eq"
  | Cne => "ne"
  | Clt => "lt"
  | Cle => "le"
  | Cgt => "gt"
  | Cge => "ge"
  end.

Definition string_of_condition (cond: condition) : string :=
  match cond with
  | Ccomp c => "comp " ++ string_of_comparison c
  | Ccompu c => "compu " ++ string_of_comparison c
  | Ccompimm c i => "compimm " ++ string_of_comparison c
  | Ccompuimm c i => "compuimm " ++ string_of_comparison c
  | _ => "<condition>"
  end.

Definition string_of_operator (op: operation) : string :=
  match op with
  | Omove => "move"
  | Ointconst _ => "intconst"
  | Ofloatconst _ => "floatconst"
  | Osingleconst _ => "singleconst"
  | Oindirectsymbol _ => "indirectsymbol"
  | Ocast8signed => "(int8)"
  | Ocast8unsigned => "(uint8)"
  | Ocast16signed => "(int16)"
  | Ocast16unsigned => "(uint16)"
  | Oneg => "neg"
  | Osub => "sub"
  | Omul => "mul"
  | Omulimm _ => "mulimm"
  | Omulhs => "mulhs"
  | Omulhu => "mulhu"
  | Odiv => "div"
  | Odivu => "divu"
  | Omod => "mod"
  | Omodu => "modu"
  | Oand => "and"
  | Oandimm _ => "andimm"
  | Oor => "or"
  | Oorimm _ => "orimm"
  | Oxor => "xor"
  | Oxorimm _ => "xorimm"
  | Onot => "not"
  | Oshl => "shl"
  | Oshlimm _ => "shlimm"
  | Oshr => "shr"
  | Oshrimm _ => "shrimm"
  | Oshrximm _ => "shrximm"
  | Oshru => "shru"
  | Oshruimm _ => "shruimm"
  | Ororimm _ => "rorimm"
  | Oshldimm _ => "shldimm"
  | Olea a => "lea " ++ string_of_addressing a
  | Ocmp c => string_of_condition c
  | Onegf => "negf"
  | Oabsf => "absf"
  | Oaddf => "addf"
  | Osubf => "subf"
  | Omulf => "mulf"
  | Odivf => "divf"
  | Onegfs => "negfs"
  | Oabsfs => "absfs"
  | Oaddfs => "addfs"
  | Osubfs => "subfs"
  | Omulfs => "mulfs"
  | Odivfs => "divfs"
  | Osingleoffloat => "singleoffloat"
  | Ofloatofsingle => "floatofsingle"
  | Ointoffloat => "intoffloat"
  | Ofloatofint => "floatofint"
  | Ointofsingle => "intofsingle"
  | Osingleofint => "singleofint"
  | Omakelong => "makelong"
  | Olowlong => "lowlong"
  | Ohighlong => "highlong"
  end.

Definition error_of_external (ef: external_function) : Errors.errmsg :=
match ef with
| EF_external i _ => Errors.MSG "external " :: Errors.CTX i :: nil
| EF_builtin i _ => Errors.MSG "builtin " :: Errors.CTX i :: nil
| EF_vload _ => Errors.msg "vload"
| EF_vstore _ => Errors.msg "vstore"
| EF_vload_global _ i _ => Errors.MSG "vload_global " :: Errors.CTX i :: nil
| EF_vstore_global _ i _ => Errors.MSG "vstore_global " :: Errors.CTX i :: nil
| EF_malloc => Errors.msg "malloc"
| EF_free => Errors.msg "free"
| EF_memcpy _ _ => Errors.msg "memcpy"
| EF_annot i _ => Errors.MSG "annot " :: Errors.CTX i :: nil
| EF_annot_val i _ => Errors.MSG "annot_val " :: Errors.CTX i :: nil
| EF_inline_asm _ => Errors.msg "inline_asm"
end.

Fixpoint hex_of_pos' (p: positive) acc :=
  match p with
  | x~0~0~0~0 => hex_of_pos' x (String "0" acc)
  | x~0~0~0~1 => hex_of_pos' x (String "1" acc)
  | x~0~0~1~0 => hex_of_pos' x (String "2" acc)
  | x~0~0~1~1 => hex_of_pos' x (String "3" acc)
  | x~0~1~0~0 => hex_of_pos' x (String "4" acc)
  | x~0~1~0~1 => hex_of_pos' x (String "5" acc)
  | x~0~1~1~0 => hex_of_pos' x (String "6" acc)
  | x~0~1~1~1 => hex_of_pos' x (String "7" acc)
  | x~1~0~0~0 => hex_of_pos' x (String "8" acc)
  | x~1~0~0~1 => hex_of_pos' x (String "9" acc)
  | x~1~0~1~0 => hex_of_pos' x (String "A" acc)
  | x~1~0~1~1 => hex_of_pos' x (String "B" acc)
  | x~1~1~0~0 => hex_of_pos' x (String "C" acc)
  | x~1~1~0~1 => hex_of_pos' x (String "D" acc)
  | x~1~1~1~0 => hex_of_pos' x (String "E" acc)
  | x~1~1~1~1 => hex_of_pos' x (String "F" acc)
  | 15 => String "F" acc
  | 14 => String "E" acc
  | 13 => String "D" acc
  | 12 => String "C" acc
  | 11 => String "B" acc
  | 10 => String "A" acc
  |  9 => String "9" acc
  |  8 => String "8" acc
  |  7 => String "7" acc
  |  6 => String "6" acc
  |  5 => String "5" acc
  |  4 => String "4" acc
  |  3 => String "3" acc
  |  2 => String "2" acc
  |  1 => String "1" acc
  end%positive.

Definition hex_of_pos p := hex_of_pos' p "".

Definition hex_of_N (n: N) : string :=
  match n with
  | N0 => "0"
  | Npos p => hex_of_pos p
  end.

Definition hex_of_Z (z: Z) : string :=
  match z with
  | Z0 => "0"
  | Zpos p => hex_of_pos p
  | Zneg n => String "-" (hex_of_pos n)
  end.

Definition hex_of_int := hex_of_Z ∘ Int.unsigned.

Definition string_of_instruction (i: instruction) : string :=
match i with
| Inop n => "goto " ++ hex_of_pos n
| Iop _ _ _ _ => "op"
| Iload κ addr _ _ _ => "load " ++ string_of_chunk κ ++ " " ++ string_of_addressing addr
| Istore κ addr _ _ _ => "store " ++ string_of_chunk κ ++ " " ++ string_of_addressing addr
| Icall _ _ _ _ _ => "call"
| Itailcall _ _ _ => "tail-call"
| Ibuiltin _ _ _ _ => "built-in"
| Icond _ _ if_so if_not => "cond _ ? " ++ hex_of_pos if_so ++ " : " ++ hex_of_pos if_not
| Ijumptable _ _ => "jumptable"
| Ireturn _ => "ret"
end.
