Require Import Bool.
Require Import MoreString MoreLib MoreMaps.
Require Import LinearCombination AbInt.
Require Import Interval.
Require Import LargePrime.
Import String Sumbool.
Import Monads List RelationClasses EquivDec Morphisms ZArith.
Import Maps Integers.
Import RField.
Require Import Interval.

(* Chose LCList or LCMap to implement linear combinations. *)
Require LCList.
Module LC <: LinearCombinationType := LCList.LC.

Set Implicit Arguments.
Unset Elimination Schemes.

Local Open Scope Z_scope.

(* Concrete QAP.

Wires, Roots, and three polynomial families.
*)

Module M := PTree.

Section MAP.

Context f feq fequiv fdec fplus fzero fopp fmult fone finv (F: @Field f feq fequiv fdec fplus fzero fopp fmult fone finv).
Context (f_of_z : Z → f).
Local Open Scope algebra_scope.

Definition Lc := LC.t f.

Let poly := M.t Lc.

Let of_int : int → Z := Int.unsigned.

Let not_too_large (r: interval) : bool := (Interval.size r <? Npos large_prime)%N.

Inductive bank := Bank_Shared (range: interval) | Bank_Local.

Definition wire_meta : Type := bank.

(* A binary representation is a list of linear combinations in the
   range 〚0;1〛, in little-endian. *)
Definition binary_repr : Type := list Lc.

(* To each variable, we associate either a linear combination and its
   range, or a binary representation. Using directly the linear
   combination associated to a binary representation would make it
   tricky to access the iᵗʰ bit. *)
Definition value := ((Lc * interval) + binary_repr)%type.

Definition sup_of_bit_numbers n := (two_power_nat n) - 1.
Definition range_of_bit_numbers n :=
  {|imin := 0; imax := sup_of_bit_numbers n|}.
Definition range_of_binary_repr (bs:binary_repr) :=
  range_of_bit_numbers (List.length bs).

Definition range_of_value (v:value) : interval :=
  match v with
  | inl (_,r) => r
  | inr bs => range_of_binary_repr bs
  end.

Definition range_invariant (v:value) : bool :=
  match v with
  | inl (_,r) => not_too_large r
  | inr bs => (Z.of_nat (List.length bs)) <? 254 (* TODO: do not hard-code this value *)
  end.

Record cqap : Type :=
{ next_wire: positive
; next_root: positive
; next_var: positive

; cV : poly
; cW : poly
; cY : poly
; meta_data: M.t wire_meta

; env: M.t value

; bcache : M.t binary_repr

; wf : (∀ w, (next_wire <= w)%positive → meta_data ! w = None)
 ∧ (∀ v, (next_var <= v)%positive → env ! v = None)
; ranges_invariant: PTree_Properties.for_all env (λ _, range_invariant)
; md_invariant: (xH < next_wire)%positive ∧ meta_data ! xH = Some (Bank_Shared (Interval.const 1))
}.

Lemma set_wf {A} x e (y: A) :
  (∀ x', (x <= x')%positive → e ! x' = None) →
  ∀ x', (Pos.succ x <= x')%positive → (M.set x y e) ! x' = None.
Proof.
  intros H x' H0. rewrite M.gso. apply H. Psatz.lia. Psatz.lia.
Qed.
Arguments set_wf {A} [x e y] _ _ _.

Lemma next_wf {A} x (e: M.t A) :
  (∀ x', (x <= x')%positive → e ! x' = None) →
  ∀ x', (Pos.succ x <= x')%positive → e ! x' = None.
Proof.
  intros H x' H0. apply H. Psatz.lia.
Qed.
Arguments next_wf {A} [x e] _ _ _.

Definition fold_equations {A B} (h: B → B → B → A) g (z: B) (q: cqap) : list A :=
  let opt_get r m := match M.get r m with Some v => v | None => (LC.zero f) end in
  let fold_lc lc := LC.fold g lc z in
  Pos.peano_rect (λ _, list A)
  nil
  (λ r acc,
   let v := opt_get r q.(cV)  in
   let w := opt_get r q.(cW) in
   let y := opt_get r q.(cY) in
   h (fold_lc v) (fold_lc w) (fold_lc y)
   :: acc
  ) q.(next_root).

Let E := Errors.errmsg.
Import STm.Notations.

Definition incr (a b: cqap) : Prop :=
  (a.(next_wire) <= b.(next_wire)
∧ a.(next_root) <= b.(next_root)
∧ a.(next_var) <= b.(next_var)
∧ (∀ r, r < a.(next_root) → M.get r a.(cV) = M.get r b.(cV))
∧ (∀ r, r < a.(next_root) → M.get r a.(cW) = M.get r b.(cW))
∧ (∀ r, r < a.(next_root) → M.get r a.(cY) = M.get r b.(cY))
∧ (∀ w, w < a.(next_wire) → M.get w a.(meta_data) = M.get w b.(meta_data))
∧ (∀ v, v < a.(next_var) → M.get v a.(env) = M.get v b.(env))
)%positive.

Instance incr_refl : Reflexive incr :=
  λ a, conj (Pos.le_refl _) (conj (Pos.le_refl _) (conj (Pos.le_refl _) (conj (λ _ _, refl_equal) (conj (λ _ _, refl_equal) (conj (λ _ _, refl_equal) (conj (λ _ _, refl_equal) (λ _ _, refl_equal))))))).

Instance incr_trans : Transitive incr.
Proof.
  intros a b c.
  unfold incr.
  intuition eauto using Pos.le_trans.
  transitivity (M.get r (cV b)); eauto using Pos.lt_le_trans.
  transitivity (M.get r (cW b)); eauto using Pos.lt_le_trans.
  transitivity (M.get r (cY b)); eauto using Pos.lt_le_trans.
  transitivity (M.get w (meta_data b)); eauto using Pos.lt_le_trans.
  transitivity (M.get v (env b)); eauto using Pos.lt_le_trans.
Qed.

Lemma fresh_wire_md w m md :
  (1 < w)%positive ∧ m ! 1 = Some (Bank_Shared (Interval.const 1)) →
   (1 < Pos.succ w)%positive ∧ (M.set w md m) ! 1 = Some (Bank_Shared (Interval.const 1)).
Proof.
  intros H. split. Psatz.lia. rewrite M.gso. easy. Psatz.lia.
Qed.
Arguments fresh_wire_md [w m md] _.

Definition fresh_wire (md: wire_meta): STm.t incr E positive.
refine (
    STm.upd (
        λ s,
        let w := s.(next_wire) in
        (exist (incr s)
          {| next_wire := Pos.succ w ; next_root := s.(next_root) ; next_var := s.(next_var)
           ; cV := s.(cV) ; cW := s.(cW) ; cY := s.(cY)
           ; meta_data := M.set w md s.(meta_data)
           ; env := env s
           ; bcache := bcache s
           ; ranges_invariant := ranges_invariant s
           ; wf := let (a, b) := wf s in conj (set_wf a) b 
           ; md_invariant := fresh_wire_md (md_invariant s)
          |}
          _,
        w))
  ).
Proof.
  abstract
  (subst w; repeat split; simpl; try Psatz.lia;
  intros; (rewrite M.gsspec, if_dec_false; [ reflexivity | Psatz.lia ])).
Defined.

Definition fresh_root : STm.t incr E positive.
refine (
    STm.upd (
        λ s,
        let r := s.(next_root) in
        (exist (incr s)
          {| next_root := Pos.succ r ; next_wire := s.(next_wire) ; next_var := s.(next_var)
           ; cV := s.(cV) ; cW := s.(cW) ; cY := s.(cY)
           ; meta_data := meta_data s
           ; env := env s
           ; bcache := bcache s
           ; wf := wf s
           ; ranges_invariant := ranges_invariant s
           ; md_invariant := md_invariant s
          |}
          _,
        r))
  ).
Proof.
  abstract (subst r; repeat split; simpl; Psatz.lia).
Defined.

Definition fresh_eqn (v w y: Lc) : STm.t incr E positive.
  refine (
    STm.upd (
        λ s,
        let r := s.(next_root) in
        (exist (incr s)
          {| next_root := Pos.succ r ; next_wire := s.(next_wire) ; next_var := s.(next_var)
           ; cV := M.set r v s.(cV)
           ; cW := M.set r w s.(cW)
           ; cY := M.set r y s.(cY)
           ; meta_data := meta_data s
           ; env := env s
           ; bcache := bcache s
           ; wf := wf s
           ; ranges_invariant := ranges_invariant s
           ; md_invariant := md_invariant s
          |}
          _,
        r))
  ).
Proof.
  abstract (
      subst r; repeat split; simpl; try Psatz.lia;
      intros r H; rewrite M.gsspec, if_dec_false; auto; Psatz.lia).
Defined.

Lemma fresh_lc_ranges_invariant v (lc: value) e:
  PTree_Properties.for_all e (λ _, range_invariant) →
  range_invariant lc →
  PTree_Properties.for_all (M.set v lc e) (λ _, range_invariant).
Proof.
  unfold is_true. rewrite ! PTree_Properties.for_all_correct.
  intros INV K x a H.
  rewrite M.gsspec in H. destruct (Coqlib.peq x v). hsplit. simpl in *. congruence. eauto.
Qed.

Definition lc_one := LC.single xH 1.

Definition lc_of_wire w : LC.t f := LC.single w 1.
Let one_minus_wire w : LC.t f := LC.sub lc_one (LC.single w 1).

Definition lc_of_f (x: f) : Lc := if x == 0 then LC.zero _ else LC.single xH x.

Definition string_of_interval (i: interval) : string :=
  ("[" ++ hex_of_Z (imin i) ++ "; " ++ hex_of_Z (imax i) ++ "]")%string.

Inductive cQAP_err_msg_kind :=
| MSG_GET_LC
| MSG_FRESH_LC
| MSG_CMP
| MSG_CMP_EQ
| MSG_ASSERT
| MSG_ASSERT_EQ
| MSG_TO_INT
.

Definition cQAP_err_msg_type (e: cQAP_err_msg_kind) : Type :=
  match e with
  | MSG_GET_LC => positive → E
  | MSG_FRESH_LC => string → value → E
  | MSG_CMP => comparison → E
  | MSG_CMP_EQ => interval → unit → E
  | MSG_ASSERT => interval → unit → E
  | MSG_ASSERT_EQ => interval → unit → E
  | MSG_TO_INT => positive → interval → E
  end.

Definition cQAP_err_msg (e: cQAP_err_msg_kind) : cQAP_err_msg_type e :=
  match e with
  | MSG_GET_LC => λ x, Errors.MSG "qap.get_lc: bad variable " :: Errors.POS x :: nil
  | MSG_FRESH_LC => λ s x, Errors.msg ("fresh_lc: overflow in " ++ s ++ " " ++ hex_of_N (Interval.size (range_of_value x)))
  | MSG_CMP => λ c, Errors.msg ("cQAP: bad comparison " ++ string_of_comparison c)
  | MSG_CMP_EQ => λ r _, Errors.msg ("cqap.cmp.eq: overflow: " ++ string_of_interval r)
  | MSG_ASSERT => λ r _, Errors.msg ("cqap.assert: overflow: " ++ string_of_interval r)
  | MSG_ASSERT_EQ => λ r _, Errors.msg ("cqap.assert_eq: overflow: " ++ string_of_interval r)
  | MSG_TO_INT => λ x r, Errors.MSG "cqap.to_int: maybe not an int ":: Errors.POS x :: Errors.MSG (" range: " ++ string_of_interval r):: nil
 end.
Global Opaque cQAP_err_msg.

Definition fresh_lc (s: string) (lc : value) : STm.t incr E positive.
refine
  match sumbool_of_bool (range_invariant lc) with
  | left EQ =>
    STm.upd (
        λ s,
        let v := s.(next_var) in
        (exist (incr s)
          {| next_root := s.(next_root) ; next_wire := s.(next_wire) ; next_var := Pos.succ v
           ; cV := s.(cV) ; cW := s.(cW) ; cY := s.(cY)
           ; meta_data := meta_data s
           ; env := M.set v lc s.(env)
           ; bcache := bcache s
           ; wf := let (a, b) := wf s in conj a (set_wf b)
           ; ranges_invariant := fresh_lc_ranges_invariant _ _ _ (ranges_invariant s) EQ
           ; md_invariant := md_invariant s
          |}
          _,
        v))
  | right NE => STm.err _ (cQAP_err_msg MSG_FRESH_LC s lc)
  end.
Proof.
  abstract (
   subst v; repeat split; simpl; try Psatz.lia;
   intros v H; rewrite M.gsspec, if_dec_false; auto; Psatz.lia).
Defined.

Lemma set_smaller_wf {A} v x' c (bs:A) :
  (v < x')%positive →
  (∀ v0, (x' <= v0)%positive → c ! v0 = None) →
  (∀ v0, (x' <= v0)%positive → (M.set v bs c) ! v0 = None).
Proof.
  intros H1 H2 v0 Hv0. rewrite M.gso; auto.
  intro H. subst v0. Psatz.lia.
Qed.

Definition cache_binary_decomposition (v: positive) (bs: binary_repr) : STm.t incr E unit.
  refine (
    STm.upd (
        λ s,
        (exist (incr s)
          {| next_root := s.(next_root) ; next_wire := s.(next_wire) ; next_var := s.(next_var)
           ; cV := s.(cV)
           ; cW := s.(cW)
           ; cY := s.(cY)
           ; meta_data := meta_data s
           ; env := env s
           ; bcache := M.set v bs s.(bcache)
           ; wf := wf s
           ; ranges_invariant := ranges_invariant s
           ; md_invariant := md_invariant s
          |}
          _,
        tt))
  ).
Proof.
  abstract (
      repeat split; simpl; try Psatz.lia;
      intros r H; rewrite M.gsspec, if_dec_false; auto; Psatz.lia).
Defined.
Arguments cache_binary_decomposition : clear implicits.

(* Injection of a binary representation into a linear combination *)
Fixpoint lc_of_binary_repr (bs:binary_repr) : Lc :=
  match bs with
    | nil => LC.zero _
    | b::bs => LC.add b (LC.muln (f_of_z 2) (lc_of_binary_repr bs))
  end.

(* A tail-recursive version *)
(* Definition lc_of_binary_repr (bs:binary_repr) : Lc := *)
(*   fst (fold_left (fun (acc : (Lc * f)%type) b => *)
(*                     let (acc, pow2) := acc in *)
(*                     let pow2' := fmult (fplus fone fone) pow2 in *)
(*                     (LC.add acc (LC.muln pow2 b), pow2') *)
(*                  ) bs (LC.zero _, 1)). *)

Definition lc_of_value (v:value) : Lc * interval :=
  match v with
  | inl lc => lc
  | inr bs => (lc_of_binary_repr bs, range_of_binary_repr bs)
  end.

Definition value_of_lc : (Lc * interval) -> value := inl.

(* The converse function: need to generate equations, based on the
   range *)
Definition input_bit :=
  do w ← fresh_wire Bank_Local;
  let lc := lc_of_wire w in
  do _ ← fresh_eqn lc (one_minus_wire w) (LC.zero _);
    STm.ret lc.

Fixpoint fresh_binary_decomposition_aux (n:nat) (acc:binary_repr) : STm.t incr E binary_repr :=
  match n with
  | O => STm.ret acc
  | S m =>
    do lc ← input_bit;
      fresh_binary_decomposition_aux m (lc::acc)
  end.

Definition fresh_binary_decomposition (n:nat) : STm.t incr E binary_repr :=
  do res ← fresh_binary_decomposition_aux n nil;
  if (size (range_of_binary_repr res) <? Npos large_prime)%N then
    STm.ret res
  else
    STm.err _ (Errors.MSG "qap.fresh_binary_decomposition: the generated binary decomposition has a range too large" :: nil).


(* Truncation and completion to 32 bits *)

Fixpoint truncate n (start : nat → binary_repr) (bs:binary_repr)
         {struct bs} :=
  match bs with
    | nil => start n
    | b::bs =>
      match n with
        | O => nil
        | S n => b::(truncate n start bs)
      end
  end.

Fixpoint zeros n :=
  match n with
  | O => nil
  | S n => (LC.zero f)::(zeros n)
  end.

Definition truncate_32 := truncate Int.wordsize (fun _ => nil).
Definition truncate_and_complete_32 := truncate Int.wordsize zeros.

(* TODO: optimize this function *)
Definition number_of_bits (i:interval) : nat :=
  S (Z.to_nat (Z.log2 (imax i))).

(* We require only the first 32 bits *)
Fixpoint positive_binary_decomposition_aux (p:positive) (n:nat) {struct p} : binary_repr :=
  match n with
  | O => nil
  | S n =>
    match p with
    | xI p' => lc_one::(positive_binary_decomposition_aux p' n)
    | xO p' => (LC.zero f)::(positive_binary_decomposition_aux p' n)
    | xH => lc_one::nil
    end
  end.

Definition positive_binary_decomposition (p:positive) :
  binary_repr :=
  positive_binary_decomposition_aux p Int.wordsize.

Definition get_binary_decomposition (a: positive) : STm.t incr E (binary_repr * option nat) :=
  do q ← STm.get (incr := _);
  match M.get a q.(bcache) with
  | Some bs => STm.ret (bs, None)
  | None =>
    match M.get a q.(env) with
    | Some (inr bs) => STm.ret (bs, None)
    | Some (inl (lc, r)) =>
      (* TODO: optimization: try to see if lc wouldn't be, by any
         chance, a binary decomposition *)
      match Interval.is_const r with
      | Some z =>
        (* If lc is a constant known at compile-time, we can compute
           its decomposition as a bunch of Boolean constants *)
        match z with
        | Zpos p =>
          let bs := positive_binary_decomposition p in
          do _ ← cache_binary_decomposition a bs;
          STm.ret (bs, None)
        | Z0 => STm.ret (nil, None)
        | Zneg _ => STm.err _ (Errors.MSG "qap.get_binary_decomposition: handles only non-negative constants" :: nil)
        end
      | None =>
        (* Otherwise, we introduce as many variables as needed to
           represent its bits *)
        do _ ← STm.assert (sumbool_of_bool (0 <=? imin r))
          (λ _, Errors.msg ("qap.gbd: bad interval: " ++ string_of_interval r));
        let n := number_of_bits r in
        do res ← fresh_binary_decomposition n;
        let lcres := lc_of_binary_repr res in
        (* TODO: optimization: isolate the bit of low weight to avoid one equation *)
        do eq ← fresh_eqn (LC.zero _) (LC.zero _ ) (LC.sub lc lcres);
        (* Return the truncated result on 32 bits *)
        let res' := truncate_32 res in
        do _ ← cache_binary_decomposition a res';
        STm.ret (res', Some n)
      end
    | None => STm.err _ (Errors.MSG "qap.get_binary_decomposition: bad variable " :: Errors.POS a :: nil)
    end
  end.

Definition get_lc (x: positive) : STm.t incr E (Lc * interval) :=
  do q ← STm.get (incr := _);
  match M.get x q.(env) with
  (* Why doesn't Coq infer this coercion? *)
  | Some v => STm.ret (lc_of_value v)
  | None => STm.err _ (cQAP_err_msg MSG_GET_LC x)
  end.


(* Bitwise operations *)
(* Fixpoint prune_aux bs := *)
(*   match bs with *)
(*   | nil => nil *)
(*   | lc::bs' => if LC.is_zero lc then prune_aux bs' else bs *)
(*   end. *)

(* Definition prune bs := List.rev (prune_aux (List.rev bs)). *)


Fixpoint bitwise_first_nil (op:Lc → Lc → STm.t incr E Lc) (y:binary_repr) {struct y} :=
  match y with
  | nil => STm.ret nil
  | lc2::y =>
      do res ← op (LC.zero _) lc2;
      do ress ← bitwise_first_nil op y;
      STm.ret (res::ress)
  end.


Fixpoint bitwise_aux (op:Lc → Lc → STm.t incr E Lc) (x y:binary_repr) {struct x} :=
  match x with
  | nil => bitwise_first_nil op y
  | lc1::x => match y with
              | nil =>
                  do res ← op lc1 (LC.zero _);
                  do ress ← bitwise_aux op x nil;
                  STm.ret (res::ress)
              | lc2::y =>
                  do res ← op lc1 lc2;
                  do ress ← bitwise_aux op x y;
                  STm.ret (res::ress)
              end
  end.


(* TODO: add pruning *)
Definition bitwise op x y :=
  (* do res ← bitwise_aux op x y; *)
  (* STm.ret (prune res. *)
  bitwise_aux op x y.

Definition bitNot a := LC.sub lc_one a.

Definition bitAnd x y :=
  match LC.is_single 1%positive x with
  | Some i => STm.ret (LC.muln i y)
  | None =>
    match LC.is_single 1%positive y with
    | Some j => STm.ret (LC.muln j x)
    | None =>
      do w ← fresh_wire Bank_Local;
      let res := lc_of_wire w in
      do r ← fresh_eqn x y res;
      STm.ret res
    end
  end.

Definition bitOr a b :=         (* a + b - ab *)
  do aux ← bitAnd a b;
  STm.ret (LC.add a (LC.sub b aux)).

Definition bitXor a b :=        (* a + b - 2ab *)
  do aux ← bitAnd a b;
  STm.ret (LC.add a (LC.sub b (LC.muln (1+1) aux))).

Definition bs_land := bitwise bitAnd.
Definition bs_lor := bitwise bitOr.
Definition bs_lxor := bitwise bitXor.

(* bs_lnot does not need to be in the monad since we do not add equation *)
(* TODO: add pruning *)
Definition bs_lnot bs := List.map bitNot (truncate_and_complete_32 bs).

Fixpoint bs_shru_aux (bs:binary_repr) n {struct n} : binary_repr :=
  match n with
    | O => bs
    | S n =>
      match bs with
        | nil => nil
        | _::bs => bs_shru_aux bs n
      end
  end.

(* >> does not commute with truncation, which is why we truncate first

   TODO: A better solution would be to ensure that
   [get_binary_decomposition] always returns a list whose length is less
   than 32, which requires to prove the invariant that all binary
   operations return such outputs on such inputs.
 *)
Definition bs_shru bs i := bs_shru_aux bs (Z.to_nat (Int.unsigned i)).

Definition get_range (a: positive) : STm.t incr E interval :=
  do q ← STm.get (incr := _);
  match
    match M.get a q.(env) with
    | Some (inl (_, r)) => Some r
    | Some (inr bs) => Some (range_of_binary_repr bs)
    | None => None
    end
  with
  | Some r => STm.ret r
  | None => STm.err _ (cQAP_err_msg MSG_GET_LC a)
  end.

Definition ceq_or_cne (c: comparison) : bool :=
  match c with Ceq | Cne => true | _ => false end.
Global Opaque ceq_or_cne.

(* Argument [c] should be Ceq or Cne only *)
Definition int_of_cbool (c: comparison) (b: bool) : int :=
  if match c with Ceq => b | _ => negb b end then Int.one else Int.zero.
Global Opaque int_of_cbool.

Definition two_const_ranges (x y: interval) :
  { ij | is_const x = Some (fst ij) ∧ is_const y = Some (snd ij) } + { True } :=
  match is_const x as c return { ij | c = Some (fst ij) ∧ is_const y = Some (snd ij) } + { True } with
  | Some i => match is_const y as d return { ij | Some i = Some (fst ij) ∧ d = Some (snd ij) } + { True } with Some j => inleft (exist _ (i, j) (conj eq_refl eq_refl)) | None => inright I end
  | None => inright I
  end.

Definition cpair (c: comparison) (x y: Lc) : Lc * Lc :=
  match c with Ceq => (x, y) | _ => (y, x) end.

Definition cQAP : Arith.t E positive incr.
  (* For arithmetic operations, we use the standard operations and
     eventually lose a binary decomposition. For bitwise operations,
     we ask for a binary decomposition, and keep it. *)
refine
  (* Why doesn't Coq infer this coercion too? *)
{| Arith.of_int i := let z := of_int i in fresh_lc "of_int" (value_of_lc (lc_of_f (f_of_z z), Interval.const z))
 ; Arith.to_int x := do r ← get_range x;
   match Interval.is_const r with
   | Some i => STm.ret (Int.repr i)
   | None => STm.err _ (cQAP_err_msg MSG_TO_INT x r)
   end
 ; Arith.add a b :=
   do x ← get_lc a; do y ← get_lc b;
   fresh_lc "add" (value_of_lc (LC.add (fst x) (fst y), Interval.add (snd x) (snd y)))
 ; Arith.sub a b :=
   do x ← get_lc a; do y ← get_lc b;
   fresh_lc "sub" (value_of_lc (LC.sub (fst x) (fst y), Interval.sub (snd x) (snd y)))
 ; Arith.mul a b :=
   do x ← get_lc a; do y ← get_lc b;
   let '(x, xr) := x in let '(y, yr) := y in
   match Interval.is_const xr with
   (* TODO: possible optimization: when multiplying a binary decomposition by a constant which is a power of 2, keep this representation *)
   | Some i => do v ← fresh_lc "mul-ny" (value_of_lc (LC.muln (f_of_z i) y, Interval.muln i yr)); STm.ret (v, false)
   | None =>
   match Interval.is_const yr with
   | Some j => do v ← fresh_lc ("mul-nx " ++ hex_of_Z j ++ " ") (value_of_lc (LC.muln (f_of_z j) x, Interval.muln j xr)); STm.ret (v, false)
   | None =>
     let range := Interval.mul xr yr in (* could be more precise ! *)
     do w ← fresh_wire Bank_Local;
     let res := lc_of_wire w in
     do r ← fresh_eqn x y res;
     do lc ← fresh_lc "mul" (value_of_lc (res, range));
     STm.ret (lc, true)
   end
   end

 ; Arith.land a b :=
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     do y_hy ← get_binary_decomposition b;
     let '(y, hy) := y_hy in
     do res ← bs_land x y;
     do res' ← fresh_lc "land" (inr res);
     STm.ret (res',(hx, hy))
 ; Arith.lor a b :=
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     do y_hy ← get_binary_decomposition b;
     let '(y, hy) := y_hy in
     do res ← bs_lor x y;
     do res' ← fresh_lc "lor" (inr res);
     STm.ret (res',(hx, hy))
 ; Arith.lxor a b :=
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     do y_hy ← get_binary_decomposition b;
     let '(y, hy) := y_hy in
     do res ← bs_lxor x y;
     do res' ← fresh_lc "lxor" (inr res);
     STm.ret (res',(hx, hy))
 (*
 ; Arith.lnot a :=
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     do res ← fresh_lc "lnot" (inr (bs_lnot x));
     STm.ret (res,hx)
  *)
     (*
 ; Arith.shl a i :=
   (* TODO: optimization: keep binary representation when it is
      already known *)
   do x ← get_lc a;
   let '(x, xr) := x in
   let m := two_p (Int.unsigned i) in
   do v ← fresh_lc (value_of_lc (LC.muln (f_of_z m) x, Interval.muln m xr));
   STm.ret v
   *)
 ; Arith.shru a i :=
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     do res' ← fresh_lc "shru" (inr (bs_shru x i));
     STm.ret (res',hx)

 ; Arith.init :=
     {| next_wire := 2 ; next_root := 1 ; next_var := 1
      ; cV := (M.empty _) ; cW := (M.empty _) ; cY := (M.empty _)
      ; meta_data := M.set xH (Bank_Shared (Interval.const 1)) (M.empty _)
      ; env := M.empty _
      ; bcache := M.empty _
      ; wf := conj _ (λ _ _, M.gempty _ _)
      ; ranges_invariant := eq_refl
      ; md_invariant := conj eq_refl eq_refl
     |}

 ; Arith.input i r :=
     match i with
     | InputNow _ i => let z := of_int i in fresh_lc "input-now" (value_of_lc (lc_of_f (f_of_z z), Interval.const z))
     | InputLater =>
       match r with
         | Some (l, u) =>
           (* Range is known: public input on one wire. *)
           let range := {| imin := Int.signed l; imax := Int.signed u |} in
           do w ← fresh_wire (Bank_Shared range);
           fresh_lc "input-later" (value_of_lc (lc_of_wire w, range))
         | None =>
           (* No range information: private input on one bit. *)
           do lc ← input_bit;
           fresh_lc "input-bit" (value_of_lc (lc, Interval.bit))
       end
     end

 ; Arith.output a :=
   do r ← get_range a;
   do lc_hx ←
   if Interval.leb r (range_of_bit_numbers Int.wordsize)
   then do lc ← get_lc a; STm.ret (lc, None)
   else
     (* Performs a truncation to 32 bits (and a binary decomposition if
        needed) *)
     do x_hx ← get_binary_decomposition a;
     let '(x, hx) := x_hx in
     let x' := truncate_32 x in
     let i := range_of_binary_repr x' in
     STm.ret (lc_of_binary_repr x', i, hx);
   let '((lc, r), hx) := lc_hx in
   do w ← fresh_wire (Bank_Shared r);
   let res := lc_of_wire w in
   do _ ← fresh_eqn (LC.zero _) (LC.zero _) (LC.sub lc res);
   do res' ← fresh_lc "output" (value_of_lc (res, r));
   STm.ret (res', hx)

; Arith.cmp s c a b :=
    if ceq_or_cne c
    then
      do x ← get_lc a; do y ← get_lc b;
      match two_const_ranges (snd x) (snd y) with
      | inleft (exist (i, j) _) =>
        let ceq := Int.eq (Int.repr i) (Int.repr j) in
        let res := of_int (int_of_cbool c ceq) in
        fresh_lc "cmp" (value_of_lc (lc_of_f (f_of_z res), Interval.const res))
      | inright _ =>
      (* w' holds the inverse of (x − y) *)
      (* w is a bit holding the result *)
      let diff := LC.sub (fst x) (fst y) in
      let range := Interval.sub (snd x) (snd y) in
      do _ ← STm.assert (sumbool_of_bool (Interval.leb range {| imin := - Int.max_unsigned; imax := Int.max_unsigned |})) (cQAP_err_msg MSG_CMP_EQ range);
      do w ← fresh_wire Bank_Local;
      do w' ← fresh_wire Bank_Local;
      let wlc := lc_of_wire w in
      let nwlc := one_minus_wire w in
      let '(α, β) := cpair c wlc nwlc in
      do r ← fresh_eqn α diff (LC.zero _);
      do r' ← fresh_eqn (lc_of_wire w') diff β;
      fresh_lc "cmp" (value_of_lc (wlc, Interval.bit))
      end
    else
      STm.err _ (cQAP_err_msg MSG_CMP c)

; Arith.assert c :=
    do xr ← get_lc c;
    let '(x, r) := xr in
    if Interval.eqb r Interval.bit
    then
      (* If we know x is a bit, just add the equation “x = 1”; namely “0 × 0 = (1 − x)” *)
      do _ ← fresh_eqn (LC.zero _) (LC.zero _) (LC.sub lc_one x);
      STm.ret None
    else
    (* Subtract Int.modulus to fit in the range *)
    (* Very ad-hoc hack *)
      let '(x, r, s) :=
        if Int.max_signed <? r.(imin)
          then (LC.sub x (LC.single xH (f_of_z Int.modulus)), Interval.sub r (Interval.const Int.modulus), S O)
          else (x, r, O)
            in
      do _ ← STm.assert
        (sumbool_of_bool (Interval.leb r {| imin := - Int.max_unsigned; imax := Int.max_unsigned |}))
        (cQAP_err_msg MSG_ASSERT r);
      (* Add equation “x · y = 1” *)
      do w ← fresh_wire Bank_Local;
      do _ ← fresh_eqn x (lc_of_wire w) lc_one;
      STm.ret (Some (w, s))

; Arith.assert_eq a b :=
    do xr ← get_lc a;
    do yr ← get_lc b;
    let range := Interval.sub (snd xr) (snd yr) in
    do _ ← STm.assert (sumbool_of_bool (Interval.leb range {| imin := - Int.max_unsigned; imax := Int.max_unsigned |})) (cQAP_err_msg MSG_ASSERT_EQ range);
    do _ ← fresh_eqn (fst xr) lc_one (fst yr);
    STm.ret tt    
|}.
Proof.
  abstract (intros w H; rewrite M.gso; [ apply M.gempty | Psatz.lia ]).
Defined.

End MAP.

Arguments lc_of_f {f feq fequiv _ _} _.
Arguments lc_of_value {f feq fequiv _ _ _ _ _} _.
Arguments cQAP {_ _ _ _ _ _ _ _ _} _.
Arguments input_bit {f feq fequiv _ _ _ _ _} _.
Arguments fresh_binary_decomposition_aux {_ _ _ _ _ _ _ _} _ _ _.
Arguments fresh_binary_decomposition {_ _ _ _ _ _ _ _} _ _.
Arguments lc_of_binary_repr {_ _ _ _ _ _ _ _} _.
Arguments positive_binary_decomposition_aux {_ _} _ _.
Arguments positive_binary_decomposition {_ _} _.
Arguments get_binary_decomposition {_ _ _ _ _ _ _ _ _ _} _ _.
Arguments get_lc {_ _ _ _ _ _ _ _} _ _.
Arguments truncate_32 {_} _.
Arguments truncate_and_complete_32 {_} _.
Arguments bitAnd {_ _ _ _ _ _ _} _ _ _.
Arguments bs_land {_ _ _ _ _ _ _} _ _ _.
Arguments bitOr {_ _ _ _ _ _ _ _ _} _ _ _.
Arguments bs_lor {_ _ _ _ _ _ _ _ _} _ _ _.
Arguments bitXor {_ _ _ _ _ _ _ _ _} _ _ _.
Arguments bs_lxor {_ _ _ _ _ _ _ _ _} _ _ _.
Arguments bs_lnot {_ _ _ _ _ _ _ _} _.
