Require Import Utf8 Bool List.
Require Import EquivDec.
Import Relation_Definitions RelationClasses Morphisms.

Unset Elimination Schemes.
Set Implicit Arguments.

Class Zero A := zero: A.
Class One A := one: A.
Notation "0" := zero : algebra_scope.
Notation "1" := one : algebra_scope.
Class Plus A := plus : A → A → A.
Class Opposite A := opposite : A → A.
Class Mult A := mult : A → A → A.
Class Inverse A := inverse : A → A.

Infix "+" := plus : algebra_scope.
Infix "×" := mult (at level 40, left associativity) : algebra_scope.
Notation "- a" := (opposite a) : algebra_scope.
Notation "a ⁻¹" := (inverse a) (at level 20) : algebra_scope.

Local Open Scope algebra_scope.

Class Associative A `{Equivalence A} (op: A → A → A) := associativity : ∀ a b c, op (op a b) c === op a (op b c).
Class Commutative A `{Equivalence A} (op: A → A → A) := commutativity : ∀ a b, op a b === op b a.
Class LeftIdentity A `{Equivalence A} (op: A → A → A) (a: A) := left_identity : ∀ b, op a b === b.
Class RightIdentity A `{Equivalence A} (op: A → A → A) (a: A) := right_identity : ∀ b, op b a === b.
Class LeftInverse A `{Equivalence A} op (f: A → A) z := left_inverse : ∀ x, op (f x) x === z.
Class LeftAbsorb A `{Equivalence A} (op: A → A → A) (a: A) := left_absorb : ∀ b, op a b === a.
Class LeftDistribute A `{Equivalence A} (f g: A → A → A) := left_distribute : ∀ a b c, f a (g b c) === g (f a b) (f a c).

Class Field A (e: relation A) `(EqDec _ e) `(Plus A) `(Zero A) `(Opposite A) `(Mult A) `(One A) `(Inverse A) :=
{ field_plus_proper :> Proper (e ==> e ==> e) plus
; field_opp_proper :> Proper (e ==> e) opposite
; field_mult_proper :> Proper (e ==> e ==> e) mult
; field_inv_proper :> Proper (e ==> e) inverse
; field_plus_associative :> Associative plus
; field_plus_commutative :> Commutative plus
; field_mult_associative :> Associative mult
; field_mult_commutative :> Commutative mult
; field_zero_id:> LeftIdentity plus 0
; field_opp_id:> LeftInverse plus opposite 0
; field_one_id:> LeftIdentity mult 1
; field_distr:> LeftDistribute mult plus
; field_not_trivial: 1 =/= 0
; field_inv_id : ∀ x, x =/= 0 → x ⁻¹ × x === 1
}.

Instance field_zero_right_id A `(Field A) : RightIdentity plus 0.
Proof. intros a. rewrite commutativity. apply left_identity. Qed.

Instance field_one_right_id A `(Field A) : RightIdentity mult 1.
Proof. intros a. rewrite commutativity. apply left_identity. Qed.

Definition sub A `{Plus A} `{Opposite A} (x y: A) : A := x + - y.
Definition div A `{Mult A} `{Inverse A} (x y: A) : A := x × y ⁻¹.

Infix "−" := sub (at level 50, left associativity) : algebra_scope.
Infix "÷" := div (at level 40, left associativity) : algebra_scope.

Instance field_sub_proper A e `(F: Field A e) : Proper (e ==> e ==> e) (sub (A := A)).
Proof. intros x x' Hx y y' Hy. apply field_plus_proper. exact Hx. apply field_opp_proper. exact Hy. Qed.

Instance field_sub_id A `(Field A) : RightIdentity (sub (A := A)) 0.
Proof.
  intros b.
  unfold sub. rewrite <- (right_identity (op := plus) (- 0)), left_inverse. apply right_identity.
Qed.

Section Ring.
Require Field_tac.
Require Import ZArith.
Import Ring_theory Field_theory.

Context A feq `(Field A feq).

Lemma ring_of_field : ring_theory 0 1 plus mult (sub (A := A)) opposite feq.
Proof.
split. exact field_zero_id. exact field_plus_commutative. symmetry. apply field_plus_associative. apply left_identity. exact field_mult_commutative. symmetry. apply field_mult_associative.
intros x y z. rewrite (commutativity _ z), left_distribute, ! (commutativity z). reflexivity.
reflexivity.
intros x. rewrite (commutativity x). apply left_inverse.
Qed.

Lemma field_of_field : field_theory 0 1 plus mult (sub (A := A)) opposite (div (A := A)) inverse Equivalence.equiv.
Proof.
split.
exact ring_of_field.
exact field_not_trivial.
reflexivity.
exact field_inv_id.
Qed.

Add Field relevent : field_of_field.

Global Instance mult_0_l : LeftAbsorb mult 0.
Proof. intro. ring. Qed.

Lemma field_integral: ∀ x y, x × y === 0 → x === 0 ∨ y === 0.
Proof.
  intros x y EQ.
  destruct (x == 0) as [ ? | xnz ]. now left. right.
  rewrite <- (left_identity y), <- (field_inv_id xnz), associativity, EQ, commutativity.
  apply left_absorb.
Qed.

Lemma field_zero_diff_eq : ∀ a b, a − b === 0 → a === b.
Proof.
  intros a b hz.
  assert (b === a − b + b) as X by (rewrite hz; ring).
  rewrite X. ring.
Qed.

Lemma field_zero_diff_neq : ∀ a b, a − b =/= 0 → a =/= b.
Proof.
  intros a b hz EQ. apply hz. rewrite EQ. ring.
Qed.

Lemma field_prod_nz: ∀ x y, x × y === 1 → x =/= 0.
Proof. intros x y X EQ. rewrite EQ, left_absorb in X. apply field_not_trivial. symmetry. exact X. Qed.

End Ring.

