Require Import Run.
Require Import Classical.
Import Coqlib Globalenvs Registers Events Memory Maps Integers.
Import Utf8 Relation_Definitions RelationClasses.
Import MoreLib Monads Expr AbState.

Unset Implicit Arguments.

Ltac vauto := (econstructor (eauto; econstructor (eauto ; fail))) || idtac.

Definition select_def (b: bool) old new :
  Values.Val.add
  (Values.Val.mul (Values.Vint new) (if b then Values.Vtrue else Values.Vfalse))
  (Values.Val.mul (Values.Vint old) (Values.Val.sub Values.Vtrue (if b then Values.Vtrue else Values.Vfalse))) = Values.Vint (if b then new else old).
Proof.
  destruct b; simpl; f_equal; [ | rewrite Int.add_commut ];
  rewrite ! Int.mul_one, Int.mul_zero; apply Int.add_zero.
Qed.

Section PROG.

Variable prog : RTL.program.
Let ge := Genv.globalenv prog.

Lemma expr_of_op_inv op args e:
  expr_of_op op args = Result e →
  ∀ sp rs m v,
    rtl_eval_expr ge sp rs m e tt = STm.Result v tt I ↔
    Op.eval_operation ge sp op (rs ## args) m = Some v.
Proof.
  unfold rtl_eval_expr, eval_expr.
  destruct op; simpl; intros H sp rs m v; split; intros K; hsplit;
  repeat match goal with
  | H : appcontext[ lea _ _ = _ ] |- _ => unfold lea in H
  | H : appcontext[ expr_of_cond _ _ = _ ] |- _ => unfold expr_of_cond in H
  | |- appcontext[ Genv.symbol_address ] => unfold Genv.symbol_address
  | H : appcontext[ match ?a with _ => _ end ] |- _ => destruct a eqn:?; hsplit; simpl in *; repeat subst end;
  simpl in *; hsplit; (congruence || subst; reflexivity || idtac).
  destruct (rs !! r); try easy. now unfold Values.Val.shl, Values.Val.mul; rewrite Heqb, Int.shl_mul.
  destruct (rs !! r); try easy. now unfold STm.bind, Values.Val.shl, Values.Val.mul; simpl; rewrite Heqb, <- Int.shl_mul.
Qed.

Corollary lea_inv a args e:
  lea a args = Result e →
  ∀ sp rs m v,
    rtl_eval_expr ge sp rs m e tt = STm.Result v tt I ↔
    Op.eval_addressing ge sp a (rs ## args) = Some v.
Proof. exact (expr_of_op_inv (Op.Olea a) args e). Qed.

Corollary expr_of_cond_inv cond args e:
  expr_of_cond cond args = Result e →
  ∀ sp rs m v,
    rtl_eval_expr ge sp rs m e tt = STm.Result v tt I ↔
    ∃ ob, Values.Val.of_optbool ob = v ∧
    Op.eval_condition cond (rs ## args) m = ob.
Proof.
  intros H sp rs m v.
  rewrite (expr_of_op_inv (Op.Ocmp cond) args e H sp rs m v). simpl.
  split; intros; hsplit; vauto. congruence.
Qed.

Arguments expr_of_op_inv [op args e] _ sp rs m v.
Arguments lea_inv [a args e] _ sp rs m v.
Arguments expr_of_cond_inv [cond args e] _ sp rs m v.

Section S.

Let E : Type := Errors.errmsg.

Context (ab_state: Type) (incr: relation ab_state)
        `{incr_refl: Reflexive _ incr} `{incr_trans: Transitive _ incr}
        (var: Type).
Variable abState : state_abstraction incr E var.

Lemma main_invariant c:
  ∀ P : (state ab_state var * atrace var) → Prop,
  (∀ pc x y tr pc' s' Hincr,
     P (inl pc, x, y) → step abState prog c (inl pc, x) = STm.Result (Ostep tr) (inl pc', s') Hincr → P (inl pc', s', rev_append tr y)) →
  (∀ pc x y r v s' Hincr,
     P (inl pc, x, y) → c ! pc = Some (RTL.Ireturn (Some r)) → sa_return abState (EVar r) x = STm.Result v s' Hincr →
     P (inr v, s', y)) →
  ∀ fuel,
  ∀ s,
    P s →
    ∀ v s' Hincr,
      main abState prog c fuel s = STm.Result v s' Hincr →
      P s' ∧ ∃ v q r, s' = (inr v, q, r).
Proof.
  intros P STEP RET fuel.
  induction fuel as [ | fuel IH ].
- simpl; intros; hsplit; fail.
- intros s H v s' Hincr. simpl. destruct s as (sl, sr). unfold STm.left, STm.bind.
  destruct (step abState prog c sl) as [ [ tr | ] s'' Hincr' | ? ] eqn: Hstep; simpl.
  +
    assert ((∃ pc y, sl = (inl pc, y)) ∧ (∃ pc' x, s'' = (inl pc', x))).
    {
      clear IH.
      unfold state, step, get_pc in *.
      hsplit. subst.
      Ltac local_t :=(hsplit; unfold tau_step, visible_step in *; ostep_inv; repeat subst; simpl in *; repeat (hsplit; subst); vauto).
      flatten_eq local_t.
    }
    intros K.
    revert Hstep.
    hsplit. subst. intros Hstep. eapply IH.
    2: eauto. eauto.
  +
    intros K. hsplit. subst.
    unfold step, get_pc in Hstep.
    hsplit. simpl in *. subst.
    unfold STm.left, STm.right, state in *. hsplit.
    flatten_eq local_t.
  + intros; hsplit.
Qed.

(* Sound state abstraction *)
Variable solution : ab_state → (var → Values.val) → Prop.

Variable match_istate : AbState.match_istate ab_state var.
Context (match_istate_ext: ∀ rs rs', (∀ x, rs # x = rs' # x) → ∀ ρ σ sp m, match_istate ρ σ sp rs' m → match_istate ρ σ sp rs m).

Hypothesis abState_sound : state_abstraction_spec abState prog solution match_istate.

Definition match_state ρ : RTL.state → RTL.code → Run.state ab_state var → Prop :=
  λ s c is,
  match s with
  | RTL.State nil func sp pc rs m =>
     (* reg/mem match *) match_istate ρ (snd is) sp rs m
   ∧ (* stack pointer invariant *) (∃ stk, sp = Values.Vptr stk Int.zero ∧ Mem.range_perm m stk 0 (RTL.fn_stacksize func) Cur Freeable)
   ∧ (* c is the body of func *) RTL.fn_code func = c
   ∧ (* pc agree *) inl pc = fst is
  | RTL.Returnstate nil (Values.Vint v) m => ∃ v', inr v' = fst is ∧ Values.Vint v = ρ v'
  | _ => False
  end.

Hypothesis incr_solution : ∀ σ σ', incr σ σ' → solution σ' ⊆ solution σ.
Hint Resolve incr_solution.

Definition not_stuck (s: RTL.state) : Prop := ¬ Smallstep.nostep RTL.step ge s.

Ltac t := hsplit; simpl in *; repeat subst; ostep_inv; simpl in *; hsplit; repeat subst.

Fixpoint assert_is_int_inv (e: expr reg) (s s': state _ _) hIncr :
  assert_is_int abState e s = STm.Result tt s' hIncr →
  s' = s ∧
  ∀ ρ, solution (snd s') ρ →
  ∀ sp rs m,
  match_istate ρ (snd s) sp rs m →
  ∃ i, rtl_eval_expr ge sp rs m e tt = STm.Result (Values.Vint i) tt I.
Proof.
  unfold assert_is_int.
  case e; simpl.
  - intros v.
    generalize (sa_is_int_sound abState_sound (snd s) v).
    destruct (sa_is_int _ _). 2: easy.
    intros H; specialize (H eq_refl).
    intro; hsplit. split. easy.
    intros ρ H0 sp rs m H1. specialize (H ρ H0 rs m sp H1).
    destruct (rs !! v); try easy. eauto.
  - intros [ i | | ] H; hsplit. split. reflexivity.
    intros ρ H sp rs m H0.
    eexists. reflexivity.
  - intros e0 H; hsplit.
  - intros op a b H; hsplit.
    destruct (never_fails _) eqn: Hop. 2: easy.
    generalize (assert_is_int_inv a s). unfold assert_is_int.
    destruct (is_int _ a). 2: easy.
    intros IHa. specialize (IHa _ _ eq_refl). destruct IHa as ( _ & IHa ).
    generalize (assert_is_int_inv b s). unfold assert_is_int.
    destruct (is_int _ b). 2: easy.
    intros IHb. specialize (IHb _ _ eq_refl). destruct IHb as ( _ & IHb ).
    simpl in *. hsplit.
    split. reflexivity.
    intros ρ H sp rs m H0.
    specialize (IHa ρ H sp rs m H0). specialize (IHb ρ H sp rs m H0).
    destruct IHa as ( i & IHa).
    destruct IHb as ( j & IHb).
    unfold STm.bind; simpl; rewrite IHa, IHb;
    destruct op; simpl in *; try easy;
    repeat match goal with
           | signed : bool |- _ => destruct signed; simpl
           | |- appcontext[ if ?b then _ else _ ] => destruct b; simpl
           | |- appcontext[ Values.Val.cmp ] => unfold Values.Val.cmp; simpl
           | |- appcontext[ Values.Val.cmpu ] => unfold Values.Val.cmpu; simpl
           | |- ∃ _, _ = _ => exact (ex_intro _ _ eq_refl)
           end.
Qed.

Lemma simulation_step :
  ∀ {c pc σ tr pc' σ' Hincr},
    step abState prog c (inl pc, σ) = STm.Result (Ostep tr) (inl pc', σ') Hincr →
    ∀ {ρ},
      solution σ' ρ →
      ∀ {s},
        match_state ρ s c (inl pc, σ) →
        not_stuck s →
        ∃ s',
          Smallstep.plus RTL.step ge s (concrete_trace ρ tr) s' ∧
          match_state ρ s' c (inl pc', σ').
Proof.
  Opaque Values.Val.mul.
  intros c pc σ tr pc' σ' Hincr STEP ρ Hsol s MS hProgress.
  destruct s as [ [ | stk ] func sp pc'' rs m | | ]; try now destruct MS.
  2: now simpl in MS; clear STEP; flatten t.
  simpl in MS. t.
  unfold step, set_pc', U.put, vload in STEP.
  t. t.
  destruct ((RTL.fn_code _) ! _) eqn: ?; t.
  destruct x0; repeat t;
  repeat (match goal with
  | H : sa_assign _ _ _ ?σ = STm.Result _ _ _, MS : match_istate ?ρ _ _ _ _ |- _ =>
    pose proof (sa_assign_sound abState_sound _ _ H _ Hsol _ _ _ MS); clear H
  | H : sa_forget _ _ ?σ = STm.Result _ _ _, MS : match_istate ?ρ _ _ _ _ |- _ =>
    pose proof (sa_forget_sound abState_sound H Hsol MS); clear H
  | H : sa_eval _ _ _ = STm.Result _ _ _, MS : match_istate _ _ _ _ _ |- _ =>
    pose proof (sa_eval_sound abState_sound H Hsol MS); clear H
  | H : sa_eval _ _ ?σ = STm.Result _ ?σ' _, MS : match_istate ?ρ ?σ _ _ _ |- _ =>
    let Hsol := fresh "Hsol" in
    assert (solution σ' ρ) as Hsol by eauto;
    pose proof (sa_eval_sound abState_sound H Hsol MS); clear H
  | H : sa_store _ _ _ ?σ = STm.Result _ _ _, MS : match_istate ?ρ _ _ _ _ |- _ =>
    pose proof (sa_store_sound abState_sound H Hsol MS); clear H
  | H : sa_annot_input _ _ _ _ ?σ = STm.Result _ _ _, MS : match_istate ?ρ _ _ _ _ |- _ =>
    pose proof (sa_annot_input_sound abState_sound H Hsol MS); clear H
  | H : sa_assert _ _ ?σ = STm.Result _ _ _, MS : match_istate ?ρ _ _ _ _ |- _ =>
    pose proof (sa_assert_sound abState_sound H Hsol MS); clear H
  | H : appcontext[ sa_vload ], MS : match_istate _ _ _ _ _ |- _ =>
    pose proof (sa_vload_sound abState_sound H eq_refl Hsol MS); clear H;
    flatten_eq t; try easy
  | H : appcontext[ sa_vstore ], MS : match_istate _ _ _ _ _ |- _ =>
    pose proof (sa_vstore_sound abState_sound H eq_refl Hsol MS); clear H;
    flatten_eq t; try easy
  | EVAL : rtl_eval_expr _ _ _ _ _ _ = STm.Result _ _ _, EOC: expr_of_cond _ _ = Result _ |- _ =>
    destruct (proj1 (expr_of_cond_inv EOC _ _ _ _) EVAL) as ( [ () | ] & ?); clear EVAL
  | EVAL : rtl_eval_expr _ _ _ _ _ _ = STm.Result _ _ _, LEA: lea _ _ = Result _ |- _ =>
    apply (proj1 (lea_inv LEA _ _ _ _)) in EVAL
  | EVAL : rtl_eval_expr _ _ _ _ _ _ = STm.Result _ _ _, EOO: expr_of_op _ _ = Result _ |- _ =>
    apply (proj1 (expr_of_op_inv EOO _ _ _ _)) in EVAL
  | H: assert_is_int _ ?e ?s = STm.Result _ ?s' ?h, MS : match_istate ?ρ ?s'' _ _ _ |- _ =>
    let K := fresh in
    pose proof (assert_is_int_inv e s s' h H) as [ ? K ]; clear H;
    t;
    try (
    let SOL := fresh in
    assert ( solution s'' ρ ) as SOL;
    [ now eauto | specialize (K _ SOL _ _ _ MS) ])
  | H : match ?a with _ => _ end _ = STm.Result _ _ _ |- _ => destruct a eqn: ?
  end; t);
  (discriminate || congruence || idtac);
  repeat subst;
  try (
  try (eexists; split; [ apply Smallstep.plus_one; econstructor (simpl; eauto; fail) | ]);
  simpl; auto;
  repeat (refine (conj _ _); eauto);
  fail).

  + (* store *)
    assert (∃ a' m', Op.eval_addressing ge (Values.Vptr stk Int.zero) a (rs ## l) = Some a' ∧ Mem.storev AST.Mint32 m a' (rs#r) = Some m') as a'.
    {
      apply not_all_not_ex. intros K. apply hProgress. intros tr s' STEP.
      inv STEP;
        match goal with H : _ ! _ = Some _, K : _ ! _ = Some _ |- _ => rewrite H in K; try discriminate;
                                                                       apply some_eq in K; inv K
 end.
      refine (K _ _). vauto.
    }
    destruct a' as (a' & m' & Ha' & Hm'). eapply lea_inv in Ha'. 2: eassumption.
    unfold Mem.storev in *. flatten idtac; try (now hsplit).
    specialize (H0 _ _ _ eq_refl Ha' Hm').
    eexists. split. apply Smallstep.plus_one. eapply RTL.exec_Istore; try eassumption.
    eapply lea_inv; eauto. eassumption.
    repeat (refine (conj _ _); eauto).
    eexists. split. reflexivity.
    repeat intro. eapply Mem.perm_store_1; eauto.

  + (* vload/addr *)
    simpl in *.
    match goal with
    | H: Genv.find_symbol _ ?s = Some ?b,
      K: Genv.invert_symbol _ ?b' = Some ?s |- _ =>
        assert (b = b') by (generalize (Genv.invert_find_symbol _ _ K); intuition congruence);
          subst
    end.
    eexists. split.
    apply Smallstep.plus_one. econstructor. eassumption.
    simpl. rewrite H1. econstructor.
    econstructor. eassumption. eassumption.
    vauto. vauto.

  + (* vload/ident *)
    simpl in *.
    eexists. split.
    apply Smallstep.plus_one. econstructor. eassumption.
    econstructor. eassumption.
    apply volatile_load_vol; eauto.
    repeat match goal with H : _ = _ |- _ => rewrite H; simpl; clear H end. vauto. vauto.

  + (* vstore *)
    simpl in *.
    eexists. split.
    apply Smallstep.plus_one. eapply RTL.exec_Ibuiltin. eassumption.
    eapply volatile_store_global_sem_intro. eassumption.
    apply volatile_store_vol; try assumption.
    repeat match goal with H : Values.Vint _ = _ |- _ => rewrite <- H; clear H end;
    repeat match goal with H : _ = _ |- _ => rewrite H; simpl; clear H end; vauto.
    vauto.

  + (* input *)
    simpl in *.
    eexists. split.
    apply Smallstep.plus_one. eapply RTL.exec_Ibuiltin. eassumption.
    simpl.
    match goal with |- appcontext[ EVint ?a :: EVint ?b :: nil ] =>
    change ( EVint a :: EVint b :: nil ) with (annot_eventvals (AST.AA_int a :: AST.AA_int b :: nil) nil)
    end.
    apply extcall_annot_sem_intro. vauto.
    vauto.

  + (* select / true *)
    clear Heqb.
    destruct ((RTL.fn_code _) ! n) as [ instr | ] eqn: ?; t.
    destruct instr; t.
    revert Heqo0. case (Pos.eqb_spec n0 pc'); intros; t.
    eexists. split. eexists.
    vauto. eright. vauto. left. reflexivity. reflexivity.
    repeat (refine (conj _ _); eauto).
    assert ( solution s''1 ρ ) as SOL. eauto.
    specialize (H0 _ SOL _ _ _ MS).
    destruct H0 as (j & Hj).
    apply (proj1 (expr_of_op_inv STEP1 _ _ _ _)) in Hj.
    rewrite H4 in Hj. t.
    rewrite H1 in H5.
    rewrite (select_def true) in H5.
    easy.

  + (* select / false *)
    clear Heqb.
    destruct ((RTL.fn_code _) ! n) as [ instr | ] eqn: ?; t.
    destruct instr; t.
    revert Heqo0. case (Pos.eqb_spec n0 pc'); intros; t.
    eexists. split. eexists.
    vauto. left. reflexivity.
    repeat (refine (conj _ _); eauto).
    assert ( solution s''1 ρ ) as SOL. eauto.
    specialize (H0 _ SOL _ _ _ MS).
    destruct H0 as (j & Hj).
    apply (proj1 (expr_of_op_inv STEP1 _ _ _ _)) in Hj.
    rewrite H4 in Hj. t.
    rewrite H1 in H5.
    rewrite (select_def false) in H5.
    eapply match_istate_ext. 2: eassumption.
    clear - H1. intros x.
    rewrite Regmap.gsspec. destruct peq. congruence. reflexivity.

  + (* select / stuck *)
    elim (hProgress). intros tr s' STEP. inversion STEP; congruence.

Qed.

Require Import Behaviors.

Hypothesis good_program : program_behaves (RTL.semantics prog) ⊆ not_wrong.

Theorem simulation fuel tr' v' σ' :
  run abState prog fuel = Result (tr', v', σ') →
  ∀ ρ,
    solution σ' ρ →
    ∃ v,
    ρ v' = Values.Vint v ∧
    program_behaves (RTL.semantics prog)
      (Terminates (concrete_trace ρ tr') v).
Proof.
  intros H ρ Hsol.
  apply result_bind_inv in H. destruct H as ( m₀ & H & Hm₀').
  destruct (Genv.init_mem prog) as [ m₀' | ] eqn: Hm₀. 2: exact (result_neq Hm₀' _). apply result_eq in Hm₀'. subst m₀'.
  apply result_bind_inv in H. destruct H as ( b & H & Hb').
  destruct (Genv.find_symbol _ _) as [ b' | ] eqn: Hb. 2: exact (result_neq Hb' _). apply result_eq in Hb'. subst b'.
  apply result_bind_inv in H. destruct H as ( f & H & Hf').
  destruct (Genv.find_funct_ptr _ _) as [ f' | ] eqn: Hf. 2: exact (result_neq Hf' _). apply result_eq in Hf'. subst f'.
  apply result_bind_inv in H. destruct H as ( () & H & Hi).
  destruct (AST.signature_eq _ _) as [ Hsig | ]. 2: exact (result_neq Hi _). clear Hi.
  apply result_bind_inv in H. destruct H as ( fd & H & Hfd).
  destruct f as [fd' | ?]. 2: exact (result_neq Hfd _). apply result_eq in Hfd. subst fd'.
  apply result_bind_inv in H. destruct H as ( σ₁ & H & Hσ₁).
  set (s₀ := RTL.Callstate nil (AST.Internal fd) nil m₀).
  assert (Smallstep.initial_state (RTL.semantics prog) s₀) as S₀ by vauto.
  cut (∃ v, ρ v' = Values.Vint v ∧ state_behaves (RTL.semantics prog) s₀ (Terminates (concrete_trace ρ tr') v)).
  intros (v & Hv & Run). exists v. split. exact Hv.
  apply (program_runs S₀).
  assumption.
  destruct (Mem.alloc m₀ 0 (RTL.fn_stacksize fd)) as (m₁ & stk) eqn: Hm₁.
  set (s₁ := RTL.State nil fd (Values.Vptr stk Int.zero) (RTL.fn_entrypoint fd)
                       (RTL.init_regs nil (RTL.fn_params fd)) m₁).
  assert (∀ ρ, solution σ₁ ρ → match_state ρ s₁ (RTL.fn_code fd) (inl (RTL.fn_entrypoint fd), σ₁)) as MS₁.
  {
    subst s₁. simpl. repeat split. eapply abState_sound.(sa_init_sound); try eassumption.
    econstructor. split. exact eq_refl.
    exact (λ ofs, Mem.perm_alloc_2 _ _ _ _ _ Hm₁ ofs Cur).
  }

  flatten_eq hsplit. subst.

  generalize (λ STEP RET H₀,
              main_invariant
                (RTL.fn_code fd)
                (
                  λ q,
                  let '(pc, σ, π) := q in
                  ∀ ρ,
                    solution σ ρ →
                    ∃ s,
                      Smallstep.star RTL.step ge s₁ (concrete_trace ρ (rev π)) s ∧
                      match_state ρ s (RTL.fn_code fd) (pc, σ)
                )
                STEP RET _ _ H₀ _ _ _ Heqr).

intros X. destruct X as ( X & Y).
+ clear σ' Hsol incr0 Heqr ρ. (* clear -incr_solution good_program. *)
  intros pc x y tr pc' σ' Hincr IH Hstep ρ Hsol.
  specialize (IH ρ (incr_solution _ _ (proj2 Hincr) _ Hsol)).
  destruct IH as (s & Hsteps & MS).
  destruct (simulation_step Hstep Hsol MS) as (s' & Hss' & MS'). clear Hstep.
  intros hStuck.
  refine (good_program _ (program_runs S₀ (state_goes_wrong (RTL.semantics prog) _ hStuck _))).
  eapply Smallstep.star_step. vauto. fold s₁. exact Hsteps. reflexivity.
  intros r hFinal. clear - MS hFinal. inv hFinal. simpl in MS. hsplit. congruence.
  exists s'.
  split. eapply Smallstep.star_trans; eauto using Smallstep.plus_star.
  rewrite !rev_append_rev, rev_app_distr, rev_involutive. apply List.map_app.
  exact MS'.
+ clear -abState_sound incr_solution.
  intros pc x y r v σ' Hincr IH Hpc Hret ρ Hsol.
  set (Hsol' := incr_solution _ _ Hincr _ Hsol).
  specialize (IH ρ Hsol').
  destruct IH as (s & Hstep & MS).
  destruct s as [ () f sp pc' rs m | ? | () () ? ]; try (easy || simpl in MS; hsplit; easy).
  destruct MS as (MS & (stk' & -> & R) & Hf & ?).
  simpl in *. hsplit. subst. rewrite <- Hf in Hpc.
  destruct (Mem.range_perm_free m stk' 0 _ R) as (m' & Hm').
  destruct (abState_sound.(sa_return_sound) Hret Hsol' MS) as (Hv & Hv').
  flatten_eq auto.
  eexists (RTL.Returnstate nil (rs !! r) m').
  split. eapply Smallstep.star_right. eassumption.
  2: symmetry; apply E0_right.
  change (rs # r) with (regmap_optget (Some r) Values.Vundef rs).
  vauto.  rewrite <- Hv. simpl. vauto.
+ unfold rev'; simpl.
  intros ρ₁ Hsol₁. specialize (MS₁ _ Hsol₁).
  eexists. vauto.
+ specialize (X _ Hsol).
  destruct X as (s & Hsteps & MS). hsplit. subst.
  unfold rev'; rewrite <- rev_alt.
  destruct s; simpl in MS.
  destruct stack; hsplit; easy. easy.
  destruct stack. 2: easy.
  destruct v0 as (); try easy. hsplit. subst.
  eexists. split. eauto.
  econstructor; vauto.
Qed.

End S.

End PROG.
