Require Import List Bool.
Require Import Utf8.

Require Import RField.
Require Import EquivDec.

Local Open Scope algebra_scope.


Section SAT.

  Context f feq fequiv fdec fplus fzero fopp fmult fone finv (F: @Field f feq fequiv fdec fplus fzero fopp fmult fone finv).
  Add Field relevent : (field_of_field F).

  Lemma neq_add_0 n m : n + m =/= 0 → n =/= 0 ∨ m =/= 0.
  Proof.
    destruct (n == 0) as [E|NE].
    - rewrite E, left_identity. intuition.
    - intuition.
  Qed.


  Variable var : Type.

  (* The Boolean expresses if the literal is positive or not *)
  Definition lit := (var * bool)%type.
  Definition clause := list lit.
  Definition cnf := list clause.

  Section Interp_bool.

    Variable bvar : var -> bool.
    Definition blit (l:lit) :=
      let (x,ispos) := l in
      if ispos then (bvar x) else negb (bvar x).
    Definition bclause (c:clause) := existsb (fun l => blit l) c.
    Definition bcnf (cnf:cnf) := forallb (fun c => bclause c) cnf = true.

  End Interp_bool.

  Section Interp_F.

    Variable nvar : var -> f.

    Definition neg v := 1 + (- v).
    Definition nlit (l:lit) :=
      let (x,ispos) := l in
      if ispos then (nvar x) else neg (nvar x).

    Definition nclause (c:clause) :=
      fold_left (fun a l => a + (nlit l)) c 0.
    Definition ncnf (cnf:cnf) :=
      fold_left (fun a c => a × (nclause c)) cnf 1 =/= 0.

  End Interp_F.

  (* If the QAP has a solution, then the original problem is SAT *)
  Section Correctness.

    Variable nvar : var -> f.
    Let bvar v := if (nvar v) == 0 then false else true.

    Hypothesis nvar_bit : forall v, (nvar v) === 0 ∨ (nvar v) === 1.

    Lemma correctness_lit : forall l,
      nlit nvar l =/= 0 -> blit bvar l = true.
    Proof.
      unfold nlit, blit, neg, bvar.
      intros [x [ | ]].
      - destruct equiv_dec as [E|NE]; auto.
      - destruct equiv_dec as [E|NE]; auto.
        destruct (nvar_bit x) as [H|H]; auto.
        rewrite H. intro H1. ring_simplify in H1. elim H1. reflexivity.
    Qed.

    Lemma nclause_nonzero c :
      nclause nvar c =/= 0 -> exists l, In l c /\ nlit nvar l =/= 0.
    Proof.
      unfold nclause.
      assert (H: forall acc, fold_left (fun a l => a + nlit nvar l) c acc =/= 0 -> acc =/= 0 \/ (exists l : lit, In l c /\ nlit nvar l =/= 0)); [ |intro H1; destruct (H _ H1) as [H2|H2]; auto; elim H2; auto; reflexivity].
      induction c as [ |l c IHc]; simpl; auto.
      intros acc H. destruct (IHc _ H) as [H1|[l0 [H1 H2]]].
      - apply neq_add_0 in H1. destruct H1 as [H1|H1].
        * left; auto.
        * right. exists l. intuition.
      - right. exists l0. intuition.
    Qed.

    Lemma correctness_clause c :
      nclause nvar c =/= 0 -> bclause bvar c = true.
    Proof.
      intro H. destruct (nclause_nonzero _ H) as [l [H1 H2]].
      unfold bclause. rewrite existsb_exists.
      exists l; split; auto using correctness_lit.
    Qed.

    Lemma ncnf_nonzero cnf :
      ncnf nvar cnf -> forall c, In c cnf -> nclause nvar c =/= 0.
    Proof.
      unfold ncnf.
      assert (H: forall acc, fold_left (fun a c => a × nclause nvar c) cnf acc =/= 0 -> acc =/= 0 /\ forall c : clause, In c cnf -> nclause nvar c =/= 0); [ |intro H1; now destruct (H _ H1)].
      induction cnf as [ |c cnf IHcnf]; simpl; auto.
      - intuition.
      - intros acc H. destruct (IHcnf _ H) as [H1 H2].
        assert (H3: acc =/= 0 ∧ nclause nvar c =/= 0).
        * case (acc == 0).
          + intro H3. rewrite H3, left_absorb in H1. elim H1. reflexivity.
          + intro H3; split; auto.
            case (nclause nvar c == 0); auto.
            intro H4. rewrite H4, commutativity, left_absorb in H1. elim H1. reflexivity.
        * clear H1. destruct H3 as [H1 H3]. split; auto.
        intros c0 [H4|H4]; auto.
        subst c; auto.
    Qed.

    Lemma correctness_cnf cnf :
      ncnf nvar cnf -> bcnf bvar cnf.
    Proof.
      intro H. assert (H1 := ncnf_nonzero cnf H).
      unfold bcnf. rewrite forallb_forall.
      auto using correctness_clause.
    Qed.

    Theorem correctness c
      (H:ncnf nvar c) : exists bvar, bcnf bvar c.
    Proof. exists bvar. now apply correctness_cnf. Qed.
   End Correctness.

End SAT.
